# Copyright (c) 2012-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Interface to library of synthesis tools, e.g., JTLV, gr1c"""
from __future__ import absolute_import
import copy
import logging
import pprint
import warnings

from tulip.interfaces import gr1c
from tulip.interfaces import gr1py
from tulip.interfaces import jtlv
from tulip.interfaces import omega as omega_int
try:
    from tulip.interfaces import slugs
except ImportError:
    slugs = None
from tulip.spec import GRSpec
from tulip import transys


logger = logging.getLogger(__name__)
_hl = '\n' + 60 * '-'


def _pstr(s):
    return '(' + str(s) + ')'


def _disj(set0):
    return ' || '.join([
        '(' + str(x) + ')'
        for x in set0])


def _conj(set0):
    return ' && '.join([
        '(' + str(x) + ')'
        for x in set0
        if x != ''])


def _conj_intersection(set0, set1, parenth=True):
    if parenth:
        return ' && '.join([
            '(' + str(x) + ')'
            for x in set0
            if x in set1])
    else:
        return ' && '.join([
            str(x)
            for x in set0
            if x in set1])


def _conj_neg(set0, parenth=True):
    if parenth:
        return ' && '.join([
            '!(' + str(x) + ')'
            for x in set0])
    else:
        return ' && '.join([
            '!' + str(x)
            for x in set0])


def _conj_neg_diff(set0, set1, parenth=True):
    if parenth:
        return ' && '.join([
            '!(' + str(x) + ')'
            for x in set0
            if x not in set1])
    else:
        return ' && '.join([
            '!' + str(x)
            for x in set0
            if x not in set1])


def mutex(iterable):
    """Mutual exclusion for all time."""
    iterable = filter(lambda x: x != '', iterable)
    if not iterable:
        return list()
    if len(iterable) <= 1:
        return []
    return [_conj([
        '!(' + str(x) + ') || (' + _conj_neg_diff(iterable, [x]) + ')'
        for x in iterable])]


def exactly_one(iterable):
    """N-ary xor.

    Contrast with pure mutual exclusion.
    """
    if len(iterable) <= 1:
        return [_pstr(x) for x in iterable]
    return ['(' + _disj([
        '(' + str(x) + ') && ' + _conj_neg_diff(iterable, [x])
        for x in iterable]) + ')']


def _conj_action(actions_dict, action_type, nxt=False, ids=None):
    """Return conjunct if C{action_type} in C{actions_dict}.

    @param actions_dict: C{dict} with pairs C{action_type_name : action_value}
    @type actions_dict: dict

    @param action_type: key to look for in C{actions_dict}
    @type action_type: hashable (here typically a str)

    @param nxt: prepend or not with the next operator
    @type nxt: bool

    @param ids: map C{action_value} -> value used in solver input,
        for example, for gr1c
    @type ids: dict

    @return:
        - conjunct (includes C{&&} operator) if:

            - C{action_type} in C{actions_dict}, and
            - C{action_value} is not the empty string (modeling "no constrain")

          includes next operator (C{X}) if C{nxt = True}.
        - empty string otherwise
    @rtype: str
    """
    if action_type not in actions_dict:
        return ''
    action = actions_dict[action_type]
    if ids is not None:
        action = ids[action]
    if action is '':
        return ''
    if nxt:
        return ' X' + _pstr(action)
    else:
        return _pstr(action)


def _conj_actions(actions_dict, solver_expr=None, nxt=False):
    """Conjunction of multiple action types.

    Includes solver expression substitution.
    See also L{_conj_action}.
    """
    logger.debug('conjunction of actions: ' + str(actions_dict))
    logger.debug('mapping to solver equivalents: ' + str(solver_expr))
    if not actions_dict:
        logger.debug('actions_dict empty, returning empty string\n')
        return ''
    if solver_expr is not None:
        actions = [solver_expr[type_name][action_value]
                   for type_name, action_value in actions_dict.iteritems()]
    else:
        actions = actions_dict
    logger.debug('after substitution: ' + str(actions))
    conjuncted_actions = _conj(actions)
    logger.debug('conjuncted actions: ' + str(conjuncted_actions) + '\n')
    if nxt:
        return ' X' + _pstr(conjuncted_actions)
    else:
        return _pstr(conjuncted_actions)

# duplicate states are impossible, because each networkx vertex is unique
# non-contiguous integerss for states fine: you are lossing efficiency
# - synth doesn't care about that


def iter2var(states, variables, statevar, bool_states, must):
    """Represent finite domain in GR(1).

    An integer or string variable can be used,
    or multiple Boolean variables.

    If the possible values are:

      - mutually exclusive (use_mutex == True)
      - bool actions have not been requested (bool_actions == False)

    then an integer or string variable represents the variable in GR(1).

    If all values are integers, then an integer is used.
    If all values are strings, then a string variable is used.
    Otherwise an exception is raised, unless Booleans have been requested.

    If the values are not mutually exclusive,
    then only Boolean variables can represent them.

    Suppose N possible values are defined.
    The int variable is allowed to take N+1 values.
    The additional value corresponds to all, e.g., actions, being False.

    If FTS values are integers,
    then the additional action is an int value.

    If FTS values are strings (e.g., 'park', 'wait'),
    then the additional action is 'none'.
    They are treated by C{spec} as an arbitrary finite domain.

    An option C{min_one} is internally available,
    in order to allow only N values of the variable.
    This requires that the variable takes at least one value each time.

    Combined with a mutex constraint, it yields an n-ary xor constraint.

    @param states: values of domain.
    @type states: iterable container of C{int}
        or iterable container of C{str}

    @param variables: to be augmented with integer or string variable
        or Boolean variables.

    @param statevar: name to use for integer or string valued variable.
    @type statevar: C{str}

    @param bool_states: if True, then use bool variables.
        Otherwise use integer or string valued variable.

    @return: C{tuple} of:
      - mapping from values to GR(1) actions.
        If Booleans are used, then GR(1) are the same.
        Otherwise, they map to e.g. 'act = "wait"' or 'act = 3'

      - constraints to be added to C{trans} and/or C{init} in GR(1)
    @rtype: C{dict}, C{list}
    """
    if not states:
        logger.debug('empty container, so empty dict for solver expr')
        return dict(), None
    logger.debug('mapping domain: ' + str(states) + '\n\t'
                 'to expression understood by a GR(1) solver.')
    assert must in {'mutex', 'xor', None}
    # options for modeling actions
    if must in {'mutex', 'xor'}:
        use_mutex = True
    else:
        use_mutex = False
    if must == 'xor':
        min_one = True
    else:
        min_one = False
    # no mutex -> cannot use int variable
    if not use_mutex:
        logger.debug('not using mutex: Booleans must model actions')
        bool_states = True
    logger.debug(
        'options for modeling actions:\n\t'
        'mutex: ' + str(use_mutex) + '\n\t'
        'min_one: ' + str(min_one))
    all_str = all(isinstance(x, str) for x in states)
    if bool_states:
        logger.debug('states modeled as Boolean variables')
        if not all_str:
            raise TypeError('If Boolean, all states must be strings.')
        state_ids = {x: x for x in states}
        variables.update({s: 'boolean' for s in states})
        # single action ?
        if len(mutex(state_ids.values())) == 0:
            return state_ids
        # handle multiple actions
        if use_mutex and not min_one:
            constraint = mutex(state_ids.values())[0]
        elif use_mutex and min_one:
            constraint = exactly_one(state_ids.values())[0]
        elif min_one:
            raise Exception('min_one requires mutex')
    else:
        logger.debug('states not modeled as Booleans')
        if statevar in variables:
            raise ValueError('state variable: ' + str(statevar) +
                             ' already exists in: ' + str(variables))
        all_int = all(isinstance(x, int) for x in states)
        if all_int:
            logger.debug('all states are integers')
            # extra value modeling all False ?
            if min_one:
                n = max(states)
            else:
                n = max(states) + 1
            f = lambda x: statevar + ' = ' + str(x)
            domain = (min(states), n)
            logger.debug('created solver variable: ' + str(statevar) +
                         '\n\t with domain: ' + str(domain))
        elif all_str:
            logger.debug('all states are strings')
            assert use_mutex
            f = lambda x: statevar + ' = "' + str(x) + '"'
            domain = list(states)
            if not min_one:
                domain += [statevar + 'none']

                logger.debug(
                    'domain has been extended, because all actions\n\t'
                    'could be False (constraint: min_one = False).')
        else:
            raise TypeError('Integer and string states must not be mixed.')
        state_ids = {x: f(x) for x in states}
        variables[statevar] = domain
        constraint = None
    logger.debug(
        'for tulip variable: ' + str(statevar) + '\n'
        'the map from [tulip action values] ---> '
        '[solver expressions] is:\n' + 2 * '\t' + str(state_ids))
    return state_ids, constraint


def _add_actions(constraint, init, trans):
    if constraint is None:
        return
    trans += ['X (' + constraint[0] + ')']
    init += constraint


def _fts2spec(
    fts, ignore_initial,
    statevar, actionvar=None,
    bool_states=False, bool_actions=False
):
    """Convert closed FTS to GR(1) representation."""
    raise Exception('deprecated')
    assert isinstance(fts, transys.FiniteTransitionSystem)
    aps = fts.aps
    states = fts.states
    actions = fts.actions
    sys_init = list()
    sys_trans = list()
    sys_vars = {ap: 'boolean' for ap in aps}
    action_ids, constraint = iter2var(
        actions, sys_vars, actionvar, bool_actions, fts.actions_must)
    _add_actions(constraint, sys_init, sys_trans)
    state_ids, constraint = iter2var(states, sys_vars, statevar,
                                     bool_states, must='xor')
    if constraint is not None:
        sys_trans += constraint
    sys_init += _sys_init_from_ts(states, state_ids, aps, ignore_initial)
    sys_trans += _sys_trans_from_ts(
        states, state_ids, fts.transitions,
        action_ids=action_ids)
    tmp_init, tmp_trans = _ap_trans_from_ts(states, state_ids, aps)
    sys_init += tmp_init
    sys_trans += tmp_trans
    return (sys_vars, sys_init, sys_trans)


def sys_to_spec(
    ofts, ignore_initial, statevar,
    bool_states=False, bool_actions=False
):
    """Convert transition system to GR(1) fragment of LTL.

    The attribute C{FTS.owner} defines who controls the system,
    as described next. It can take values C{'env'} or C{'sys'}.

    The following are represented by variables controlled by C{ofts.owner}:

      - the current state
      - the atomic propositions annotating states
      - the system actions annotating edges

    The following are represented by variables controlled by the other player:

      - the environment actions annotating edges

    Multiple types of environment and system actions can be defined.
    Make sure that, depending on the player,
    C{'env'} or C{'sys'} are part of the action type names,
    so that L{synth.synthesize} can recognize them.

    Caution
    =======
    There are aspects of L{FTS} that
    need to be separately specified in a logic formula.

    An example are the initial conditions constraining the values
    of environment and system actions.

    See also
    ========
    L{sys_trans_from_ts}, L{env_open_fts2spec},
    L{create_actions}, L{create_states}

    @param ofts: L{FTS}

    @param ignore_initial: Do not include initial state info from TS.
        Enable this to mask absence of FTS initial states.
        Useful when initial states are specified in another way,
        e.g., directly augmenting the spec part.
    @type ignore_initial: C{bool}

    @param state_var: name to be used for the integer or string
        variable that equals the current transition system state.
    @type state_var: C{str}

    @param bool_states: deprecated as inefficient

        if C{True}, then use one Boolean variable
        to represent each state in GR(1).
        Otherwise use a single integer variable,
        different values of which correspond to states of C{ofts}.
    @type bool_states: bool

    @param bool_actions: Similar to C{bool_states}.
        For each type of system actions,
        and each type of environment actions:

          - if C{True}, then for each possible value of that action type,
            use a different Boolean variable to represent it.

          - Otherwise use a single integer variable,
            that ranges over the possible action values.

    @return: logic formula in GR(1) form representing C{ofts}.
    @rtype: L{GRSpec}
    """
    if not isinstance(ofts, transys.FiniteTransitionSystem):
        raise TypeError('ofts must be FTS, got instead: ' + str(type(ofts)))
    assert ofts.owner == 'sys'
    aps = ofts.aps
    states = ofts.states
    trans = ofts.transitions
    # init
    sys_init = list()
    sys_trans = list()
    env_init = list()
    env_trans = list()
    sys_vars = {ap: 'boolean' for ap in aps}
    env_vars = dict()
    actions = ofts.actions
    sys_action_ids = dict()
    env_action_ids = dict()
    for action_type, codomain in actions.iteritems():
        msg = 'action_type:\n\t' + str(action_type) + '\n'
        msg += 'with codomain:\n\t' + str(codomain)
        logger.debug(msg)
        if 'sys' in action_type:
            logger.debug('Found sys action')
            action_ids, constraint = iter2var(
                codomain, sys_vars,
                action_type, bool_actions, ofts.sys_actions_must)
            _add_actions(constraint, sys_init, sys_trans)
            logger.debug('Updating sys_action_ids with:\n\t' + str(action_ids))
            sys_action_ids[action_type] = action_ids
        elif 'env' in action_type:
            logger.debug('Found env action')
            action_ids, constrait = iter2var(
                codomain, env_vars,
                action_type, bool_actions, ofts.env_actions_must)
            _add_actions(constraint, env_init, env_trans)
            logger.debug('Updating env_action_ids with:\n\t' + str(action_ids))
            env_action_ids[action_type] = action_ids
    state_ids, constraint = iter2var(states, sys_vars, statevar,
                                     bool_states, must='xor')
    if constraint is not None:
        sys_trans += constraint
    sys_init += _sys_init_from_ts(states, state_ids, aps, ignore_initial)
    sys_trans += _sys_trans_from_ts(
        states, state_ids, trans,
        sys_action_ids=sys_action_ids, env_action_ids=env_action_ids)
    tmp_init, tmp_trans = _ap_trans_from_ts(states, state_ids, aps)
    sys_init += tmp_init
    sys_trans += tmp_trans
    env_trans += _env_trans_from_sys_ts(
        states, state_ids, trans, env_action_ids)
    return GRSpec(
        sys_vars=sys_vars, env_vars=env_vars,
        env_init=env_init, sys_init=sys_init,
        env_safety=env_trans, sys_safety=sys_trans)


def env_to_spec(
    ofts, ignore_initial, statevar,
    bool_states=False, bool_actions=False
):
    """Convert env transition system to GR(1) representation.

    The following are represented by environment variables:

      - the current state
      - the atomic propositions annotating states
      - the environment actions annotating edges

    The following are represented by system variables:

      - the system actions annotating edges

    Multiple types of environment and system actions can be defined.

    For more details see L{sys_to_spec}.

    See also
    ========
    L{sys_open_fts2spec}
    """
    if not isinstance(ofts, transys.FiniteTransitionSystem):
        raise TypeError('ofts must be FTS, got instead: ' + str(type(ofts)))
    assert ofts.owner == 'env'
    aps = ofts.aps
    states = ofts.states
    trans = ofts.transitions
    # init
    sys_init = list()
    sys_trans = list()
    env_init = list()
    env_trans = list()
    # since APs are tied to env states, let them be env variables
    env_vars = {ap: 'boolean' for ap in aps}
    sys_vars = dict()
    actions = ofts.actions
    sys_action_ids = dict()
    env_action_ids = dict()
    for action_type, codomain in actions.iteritems():
        if 'sys' in action_type:
            action_ids, constraint = iter2var(
                codomain, sys_vars, action_type,
                bool_actions, ofts.sys_actions_must)
            _add_actions(constraint, sys_init, sys_trans)
            sys_action_ids[action_type] = action_ids
        elif 'env' in action_type:
            action_ids, constraint = iter2var(
                codomain, env_vars, action_type,
                bool_actions, ofts.env_actions_must)
            _add_actions(constraint, env_init, env_trans)
            env_action_ids[action_type] = action_ids
    # some duplication here, because we don't know
    # whether the user will provide a system TS as well
    # and whether that TS will contain all the system actions
    # defined in the environment TS
    state_ids, constraint = iter2var(states, env_vars, statevar,
                                     bool_states, must='xor')
    if constraint is not None:
        env_trans += constraint
    env_init += _sys_init_from_ts(states, state_ids, aps, ignore_initial)
    env_trans += _env_trans_from_env_ts(
        states, state_ids, trans,
        env_action_ids=env_action_ids, sys_action_ids=sys_action_ids)
    tmp_init, tmp_trans = _ap_trans_from_ts(states, state_ids, aps)
    env_init += tmp_init
    env_trans += tmp_trans
    return GRSpec(
        sys_vars=sys_vars, env_vars=env_vars,
        env_init=env_init, sys_init=sys_init,
        env_safety=env_trans, sys_safety=sys_trans)


def _sys_init_from_ts(states, state_ids, aps, ignore_initial=False):
    """Initial state, including enforcement of exactly one."""
    init = []
    # skip ?
    if ignore_initial:
        return init
    if not states.initial:
        msg = (
            'FTS has no initial states.\n'
            'Enforcing this renders False the GR(1):\n'
            ' - guarantee if this is a system TS,\n'
            '   so the spec becomes trivially False.\n'
            ' - assumption if this is an environment TS,\n'
            '   so the spec becomes trivially True.')
        raise Exception(msg)
        init += ['False']
        return init
    init += [_disj([state_ids[s] for s in states.initial])]
    return init


def _sys_trans_from_ts(
    states, state_ids, trans,
    action_ids=None, sys_action_ids=None, env_action_ids=None
):
    """Convert transition relation to GR(1) sys_safety.

    The transition relation may be closed or open,
    i.e., depend only on system, or also on environment actions.

    No mutexes enforced by this function among:

        - sys states
        - env actions

    An edge attribute 'previous' can be optionally set to
    an iterable of edge attribute keys.
    The actions with those action_types those keys
    will not be prepended by the next operator.

    This enables defining both current and next actions, e.g.,

    some_action && X(some_other_action)

    About label type checking: in principle everything should work the
    same if the base class LabeledDiGraph was replaced by MultiDiGraph,
    so that users can play around with their own bare graphs,
    when they don't need the label typing overhead.

    @param trans: L{Transitions} as from the transitions
        attribute of L{FTS}.

    @param action_ids: same as C{sys-action_ids}
        Caution: to be removed in a future release

    @param sys_action_ids: dict of dicts
        outer dict keyed by action_type
        each inner dict keyed by action_value
        each inner dict value is the solver expression for that action value

        for example an action type with an
        arbitrary finite discrete codomain can be modeled either:

          - as Boolean variables, so each possible action value
            becomes a different Boolean variable with the same
            name, thus C{sys_action_ids[action_type]} will be
            the identity map on C{action_values} for that C{action_type}.

          - as integer variables, so each possible action value
            becomes a different expression in the solver (e.g. gr1c)
            input format. Then C{sys_action_ids[action_type]} maps
            C{action_value} -> solver expression of the form:

            C{action_type = i}

            where C{i} corresponds to that particular  C{action_type}.

    @param env_action_ids: same as C{sys-action_ids}
    """
    logger.debug('modeling sys transitions in logic')
    sys_trans = list()
    # Transitions
    for from_state in states:
        from_state_id = state_ids[from_state]
        precond = _pstr(from_state_id)
        cur_trans = trans.find([from_state])
        msg = ('from state: ' + str(from_state) +
               ', the available transitions are:\n\t' + str(cur_trans))
        logger.debug(msg)
        # no successor states ?
        if not cur_trans:
            logger.debug('state: ' + str(from_state) + ' is deadend !')
            sys_trans += [precond + ' -> X(False)']
            continue
        cur_str = list()
        for (from_state, to_state, label) in cur_trans:
            to_state_id = state_ids[to_state]
            postcond = ['X' + _pstr(to_state_id)]
            logger.debug('label = ' + str(label))
            if 'previous' in label:
                previous = label['previous']
            else:
                previous = set()
            logger.debug('previous = ' + str(previous))
            env_actions = {k: v for k, v in label.iteritems() if 'env' in k}
            prev_env_act = {k: v for k, v in env_actions.iteritems()
                            if k in previous}
            next_env_act = {k: v for k, v in env_actions.iteritems()
                            if k not in previous}
            postcond += [_conj_actions(prev_env_act, env_action_ids,
                                       nxt=False)]
            postcond += [_conj_actions(next_env_act, env_action_ids,
                                       nxt=True)]
            sys_actions = {k: v for k, v in label.iteritems() if 'sys' in k}
            prev_sys_act = {k: v for k, v in sys_actions.iteritems()
                            if k in previous}
            next_sys_act = {k: v for k, v in sys_actions.iteritems()
                            if k not in previous}
            postcond += [_conj_actions(prev_sys_act, sys_action_ids,
                                       nxt=False)]
            postcond += [_conj_actions(next_sys_act, sys_action_ids,
                                       nxt=True)]
            # if system FTS given
            # in case 'actions in label, then action_ids is a dict,
            # not a dict of dicts, because certainly this came
            # from an FTS, not an OpenFTS
            if 'actions' in previous:
                postcond += [_conj_action(label, 'actions',
                                          ids=action_ids, nxt=False)]
            else:
                postcond += [_conj_action(label, 'actions',
                                          ids=action_ids, nxt=True)]
            cur_str += [_conj(postcond)]
            msg = (
                'guard to state: ' + str(to_state) +
                ', with state_id: ' + str(to_state_id) +
                ', has post-conditions: ' + str(postcond))
            logger.debug(msg)
        sys_trans += [precond + ' -> (' + _disj(cur_str) + ')']
    return sys_trans


def _env_trans_from_sys_ts(states, state_ids, trans, env_action_ids):
    """Convert environment actions to GR(1) env_safety.

    This constrains the actions available next to the environment
    based on the system FTS.

    Purpose is to prevent env from blocking sys by purely
    picking a combination of actions for which sys has no outgoing
    transition from that state.

    Might become optional in the future,
    depending on the desired way of defining env behavior.

    @param env_action_ids: dict of dicts, see L{sys_trans_from_ts}.
    """
    env_trans = list()
    # this probably useless for multiple action types
    if not env_action_ids:
        return env_trans
    for from_state in states:
        from_state_id = state_ids[from_state]
        precond = _pstr(from_state_id)
        cur_trans = trans.find([from_state])
        # no successor states ?
        if not cur_trans:
            # nothing modeled for env, since sys has X(False) anyway
            # for action_type, codomain_map in env_action_ids.iteritems():
            # env_trans += [precond + ' -> X(' + s + ')']
            continue
        # collect possible next env actions
        next_env_action_combs = set()
        for (from_state, to_state, label) in cur_trans:
            env_actions = {k: v for k, v in label.iteritems() if 'env' in k}
            if not env_actions:
                continue
            logger.debug('env_actions: ' + str(env_actions))
            logger.debug('env_action_ids: ' + str(env_action_ids))
            env_action_comb = _conj_actions(env_actions, env_action_ids)
            logger.debug('env_action_comb: ' + str(env_action_comb))
            next_env_action_combs.add(env_action_comb)
        next_env_actions = _disj(next_env_action_combs)
        logger.debug('next_env_actions: ' + str(next_env_actions))
        # no next env actions ?
        if not next_env_actions:
            continue
        env_trans += [precond + ' -> X(' +
                      next_env_actions + ')']
    return env_trans


def _env_trans_from_env_ts(
    states, state_ids, trans,
    action_ids=None, env_action_ids=None, sys_action_ids=None
):
    """Convert environment TS transitions to GR(1) representation.

    This contributes to the \rho_e(X, Y, X') part of the spec,
    i.e., constrains the next environment state variables' valuation
    depending on the previous environment state variables valuation
    and the previous system action (system output).
    """
    env_trans = list()
    for from_state in states:
        from_state_id = state_ids[from_state]
        precond = _pstr(from_state_id)
        cur_trans = trans.find([from_state])
        # no successor states ?
        if not cur_trans:
            env_trans += [precond + ' -> X(False)']
            msg = (
                'Environment dead-end found.\n'
                'If sys can force env to dead-end,\n'
                'then GR(1) assumption becomes False,\n'
                'and spec trivially True.')
            warnings.warn(msg)
            continue
        cur_list = list()
        found_free = False  # any environment transition
        # not conditioned on the previous system output ?
        for (from_state, to_state, label) in cur_trans:
            to_state_id = state_ids[to_state]
            postcond = ['X' + _pstr(to_state_id)]
            env_actions = {k: v for k, v in label.iteritems() if 'env' in k}
            postcond += [_conj_actions(env_actions, env_action_ids, nxt=True)]
            # remember: this is an environment FTS, so no next for sys
            sys_actions = {k: v for k, v in label.iteritems() if 'sys' in k}
            postcond += [_conj_actions(sys_actions, sys_action_ids)]
            postcond += [_conj_action(label, 'actions', nxt=True,
                                      ids=action_ids)]
            # todo: test this claus
            if not sys_actions:
                found_free = True
            cur_list += [_conj(postcond)]
        # can sys kill env by setting all previous sys outputs to False ?
        # then env assumption becomes False,
        # so the spec trivially True: avoid this
        if not found_free and sys_action_ids:
            msg = 'no free env outgoing transition found\n' +\
                  'instead will take disjunction with negated sys actions'
            logger.debug(msg)
            for action_type, codomain in sys_action_ids.iteritems():
                conj = _conj_neg(codomain.itervalues())
                cur_list += [conj]
                msg = (
                    'for action_type: ' + str(action_type) + '\n' +
                    'with codomain: ' + str(codomain) + '\n' +
                    'the negated conjunction is: ' + str(conj))
                logger.debug(msg)
        env_trans += [_pstr(precond) + ' -> (' + _disj(cur_list) + ')']
    return env_trans


def _ap_trans_from_ts(states, state_ids, aps):
    """Require atomic propositions to follow states according to label.
    """
    init = list()
    trans = list()
    # no AP labels ?
    if not aps:
        return (init, trans)
    # initial labeling
    for state in states:
        state_id = state_ids[state]
        label = states[state]
        ap_str = _sprint_aps(label, aps)
        if not ap_str:
            continue
        init += ['!(' + _pstr(state_id) + ') || (' + ap_str + ')']
    # transitions of labels
    for state in states:
        label = states[state]
        state_id = state_ids[state]
        tmp = _sprint_aps(label, aps)
        if not tmp:
            continue
        trans += ['X((' + str(state_id) + ') -> (' + tmp + '))']
    return (init, trans)


def _sprint_aps(label, aps):
    if 'ap' in label:
        tmp0 = _conj_intersection(aps, label['ap'], parenth=False)
    else:
        tmp0 = ''
    if 'ap' in label:
        tmp1 = _conj_neg_diff(aps, label['ap'], parenth=False)
    else:
        tmp1 = _conj_neg(aps, parenth=False)
    if len(tmp0) > 0 and len(tmp1) > 0:
        tmp = tmp0 + ' && ' + tmp1
    else:
        tmp = tmp0 + tmp1
    return tmp


def build_dependent_var_table(fts, statevar):
    """Return a C{dict} of substitution rules for dependent variables.

    The dependent variables in a transition system are the
    atomic propositions that are used to label states.

    They are "dependent" because their values are completely
    determined by knowledge of the current state in the
    transition system.

    The returned substitutions can be used

    @type fts: L{FTS}

    @param statevar: name of variable used for the current state
        For example if it is 'loc', then the states
        C{'s0', 's1'} are mapped to::

          {'s0': '(loc = "s0")',
           's1': '(loc = "s1")'}

    @type state_ids: C{dict}

    @rtype: C{{'p': '((loc = "s1") | (loc = "s2") | ...)', ...}}
        where:

          - C{'p'} is a proposition in C{fts.atomic_propositions}
          - the states "s1", "s2" are labeled with C{'p'}
          - C{loc} is the string variable used for the state of C{fts}.
    """
    state_ids, __ = iter2var(fts.states, variables=dict(), statevar=statevar,
                             bool_states=False, must='xor')
    ap2states = map_ap_to_states(fts)
    return {k: _disj(state_ids[x] for x in v)
            for k, v in ap2states.iteritems()}


def map_ap_to_states(fts):
    """For each proposition find the states labeled with it.

    @type fts: L{FTS}

    @rtype: C{{'p': s, ...}} where C{'p'} a proposition and
        C{s} a set of states in C{fts}.
    """
    table = {p: set() for p in fts.atomic_propositions}
    for u in fts:
        for p in fts.node[u]['ap']:
            table[p].add(u)
    return table


def synthesize_many(specs, ts=None, ignore_init=None,
                    bool_actions=None, solver='gr1c'):
    """Synthesize from logic specs and multiple transition systems.

    The transition systems are composed synchronously, i.e.,
    they all have to take a transition at each time step.
    The synchronous composition is obtained by taking the
    conjunction of the formulas describing each transition system.

    The states of each transition system can be either:

      - all integers, or
      - all strings

    In either case the transition system state will be
    represented in logic with a single variable,
    that ranges over a finite set of integers or strings, respectively.

    The keys of C{ts} are used to name each state variable.
    So the logic formula for C{ts['name']} will be C{'name'}.

    Who controls this state variable is determined from
    the attribute C{FTS.owner} that can take the values:

      - C{'env'}
      - C{'sys'}

    For example:

      >>> ts.states.add_from(xrange(4))
      >>> ts['door'].owner = 'env'

    will result in a logic formula with
    an integer variable C{'door'}
    controlled by the environment and
    taking values over C{{0, 1, 2, 3}}.

    The example:

      >>> ts.states.add_from(['a', 'b', 'c'])
      >>> ts['door'].owner = 'sys'

    will instead result in a string variable C{'door'}
    controlled by the system and taking
    values over C{{'a', 'b', 'c'}}.

    @type specs: L{GRSpec}

    @type ts: C{dict} of L{FiniteTransitionSystem}

    @type ignore_init: C{set} of keys from C{ts}

    @type bool_actions: C{set} of keys from C{ts}

    @param solver: 'gr1c' or 'slugs' or 'jtlv'
    @type solver: str
    """
    assert isinstance(ts, dict)
    for name, t in ts.iteritems():
        assert isinstance(t, transys.FiniteTransitionSystem)
        ignore = name in ignore_init
        bool_act = name in bool_actions
        statevar = name
        if t.owner == 'sys':
            specs |= sys_to_spec(t, ignore, statevar,
                                 bool_actions=bool_act)
        elif t.owner == 'env':
            specs |= env_to_spec(t, ignore, statevar,
                                 bool_actions=bool_act)
    if solver == 'gr1c':
        ctrl = gr1c.synthesize(specs)
    elif solver == 'slugs':
        if slugs is None:
            raise ValueError('Import of slugs interface failed. ' +
                             'Please verify installation of "slugs".')
        ctrl = slugs.synthesize(specs)
    elif solver == 'jtlv':
        ctrl = jtlv.synthesize(specs)
    else:
        raise Exception('Unknown solver: ' + str(solver) + '. '
                        'Available solvers: "jtlv", "gr1c", and "slugs"')
    try:
        logger.debug('Mealy machine has: n = ' +
                     str(len(ctrl.states)) + ' states.')
    except:
        logger.debug('No Mealy machine returned.')
    # no controller found ?
    # counterstrategy not constructed by synthesize
    if not isinstance(ctrl, transys.MealyMachine):
        return None
    ctrl.remove_deadends()
    return ctrl


def synthesize(
    option, specs, env=None, sys=None,
    ignore_env_init=False, ignore_sys_init=False,
    rm_deadends=True
):
    """Function to call the appropriate synthesis tool on the specification.

    There are three attributes of C{specs} that define what
    kind of controller you are looking for:

    1. C{moore}: What information the controller knows when deciding the next
       values of controlled variables:
        - Moore: can read current state,
          but not next environment variable values, or
        - Mealy: can read current state and next environment variable values.

    2. C{qinit}: Quantification of initial variable values:
        Whether all states that satisfy a predicate should be winning,
        or the initial values of some (or all) the variables is
        subject to the synthesizer's choice.

    3. C{plus_one}: The form of assume-guarantee specification,
        i.e., how the system guarantees relate to assumptions about the
        environment.

    For more details about these attributes, see L{GRSpec}.

    The states of the transition system can be either:

      - all integers, or
      - all strings

    For more details of how the transition system is represented in
    logic look at L{synthesize_many}.

    Beware!
    =======
    This function provides a generic interface to a variety
    of routines.  Being under active development, the types of
    arguments supported and types of objects returned may change
    without notice.

    @param option: Magic string that declares what tool to invoke,
        what method to use, etc.  Currently recognized forms:

        For GR(1) synthesis:

          - C{"gr1c"}: use gr1c via L{interfaces.gr1c}.
            written in C using CUDD, symbolic

          - C{"gr1py"}: use gr1py via L{interfaces.gr1py}.
            Python, enumerative

          - C{"omega"}: use omega via L{interfaces.omega}.
            Python using C{dd} or Cython using CUDD, symbolic

          - C{"slugs"}: use slugs via L{interfaces.slugs}.
            C++ using CUDD, symbolic

          - C{"jtlv"}: use JTLV via L{interfaces.jtlv}.
            Java, symbolic
            (deprecated)

    @type specs: L{spec.GRSpec}

    @param env: A transition system describing the environment:

            - states controlled by environment
            - input: sys_actions
            - output: env_actions
            - initial states constrain the environment

        This constrains the transitions available to
        the environment, given the outputs from the system.
    @type env: L{FTS}

    @param sys: A transition system describing the system:

            - states controlled by the system
            - input: env_actions
            - output: sys_actions
            - initial states constrain the system

    @type sys: L{FTS}

    @param ignore_env_init: Ignore any initial state information
        contained in env.
    @type ignore_env_init: bool

    @param ignore_sys_init: Ignore any initial state information
        contained in sys.
    @type ignore_sys_init: bool

    @param rm_deadends: return a strategy that contains no terminal states.
    @type rm_deadends: bool

    @return: If spec is realizable,
        then return a Mealy machine implementing the strategy.
        Otherwise return None.
    @rtype: L{MealyMachine} or None
    """
    specs = _spec_plus_sys(
        specs, env, sys,
        ignore_env_init,
        ignore_sys_init)
    if option == 'gr1c':
        strategy = gr1c.synthesize(specs)
    elif option == 'slugs':
        if slugs is None:
            raise ValueError('Import of slugs interface failed. ' +
                             'Please verify installation of "slugs".')
        strategy = slugs.synthesize(specs)
    elif option == 'gr1py':
        strategy = gr1py.synthesize(specs)
    elif option == 'omega':
        strategy = omega_int.synthesize_enumerated_streett(specs)
    elif option == 'jtlv':
        strategy = jtlv.synthesize(specs)
        if isinstance(strategy, list):
            # Discard counter-examples, because here we only care that
            # it is not realizable.
            strategy = None
    else:
        raise Exception('Undefined synthesis option. ' +
                        'Current options are "gr1c", ' +
                        '"slugs", "gr1py", "omega", and "jtlv".')

    # While the return values of the solver interfaces vary, we expect
    # here that strategy is either None to indicate unrealizable or a
    # networkx.DiGraph ready to be passed to strategy2mealy().
    if strategy is None:
        return None

    ctrl = strategy2mealy(strategy, specs)
    logger.debug('Mealy machine has: n = ' +
                 str(len(ctrl.states)) + ' states.')

    if rm_deadends:
        ctrl.remove_deadends()
    return ctrl


def is_realizable(
    option, specs, env=None, sys=None,
    ignore_env_init=False, ignore_sys_init=False
):
    """Check realizability.

    For details see L{synthesize}.
    """
    specs = _spec_plus_sys(
        specs, env, sys,
        ignore_env_init, ignore_sys_init)
    if option == 'gr1c':
        r = gr1c.check_realizable(specs)
    elif option == 'slugs':
        if slugs is None:
            raise ValueError('Import of slugs interface failed. ' +
                             'Please verify installation of "slugs".')
        r = slugs.check_realizable(specs)
    elif option == 'gr1py':
        r = gr1py.check_realizable(specs)
    elif option == 'omega':
        r = omega_int.is_realizable(specs)
    elif option == 'jtlv':
        r = jtlv.check_realizable(specs)
    else:
        raise Exception('Undefined synthesis option. ' +
                        'Current options are "jtlv", "gr1c", ' +
                        '"slugs", and "gr1py"')
    if r:
        logger.debug('is realizable')
    else:
        logger.debug('is not realizable')
    return r


def _spec_plus_sys(
    specs, env, sys,
    ignore_env_init, ignore_sys_init
):
    if sys is not None:
        if hasattr(sys, 'state_varname'):
            statevar = sys.state_varname
        else:
            logger.info('sys.state_varname undefined. '
                        'Will use the default variable name: "loc".')
            statevar = 'loc'
        sys_formula = sys_to_spec(
            sys, ignore_sys_init,
            bool_states=False,
            bool_actions=False,
            statevar=statevar)
        # consider sys just a formula,
        # not a synthesis problem
        # so overwrite settings
        if hasattr(sys, 'moore'):
            cp = sys
        else:
            cp = specs
        sys_formula.moore = cp.moore
        sys_formula.plus_one = cp.plus_one
        sys_formula.qinit = cp.qinit
        specs = specs | sys_formula
        logger.debug('sys TS:\n' + str(sys_formula.pretty()) + _hl)
    if env is not None:
        if hasattr(env, 'state_varname'):
            statevar = sys.state_varname
        else:
            logger.info('env.state_varname undefined. '
                        'Will use the default variable name: "eloc".')
            statevar = 'eloc'
        env_formula = env_to_spec(
            env, ignore_env_init,
            bool_states=False,
            bool_actions=False,
            statevar=statevar)
        if hasattr(env, 'moore'):
            cp = env
        else:
            cp = specs
        env_formula.moore = cp.moore
        env_formula.plus_one = cp.plus_one
        env_formula.qinit = cp.qinit
        specs = specs | env_formula
        logger.debug('env TS:\n' + str(env_formula.pretty()) + _hl)
    logger.info('Overall Spec:\n' + str(specs.pretty()) + _hl)
    return specs


def strategy2mealy(A, spec):
    """Convert strategy to Mealy transducer.

    Note that the strategy is a deterministic game graph,
    but the input C{A} is given as the contraction of
    this game graph.

    @param A: strategy
    @type A: C{networkx.DiGraph}

    @type spec: L{GRSpec}

    @rtype: L{MealyMachine}
    """
    assert len(A) > 0
    logger.info('converting strategy (compact) to Mealy machine')
    env_vars = spec.env_vars
    sys_vars = spec.sys_vars
    mach = transys.MealyMachine()
    inputs = transys.machines.create_machine_ports(env_vars)
    mach.add_inputs(inputs)
    outputs = transys.machines.create_machine_ports(sys_vars)
    mach.add_outputs(outputs)
    str_vars = {
        k: v for k, v in env_vars.iteritems()
        if isinstance(v, list)}
    str_vars.update({
        k: v for k, v in sys_vars.iteritems()
        if isinstance(v, list)})
    mach.states.add_from(A)
    # transitions labeled with I/O
    for u in A:
        for v in A.successors_iter(u):
            d = A.node[v]['state']
            d = _int2str(d, str_vars)
            mach.transitions.add(u, v, **d)

            logger.info('node: {v}, state: {d}'.format(v=v, d=d))
    # special initial state, for first reaction
    initial_state = 'Sinit'
    mach.states.add(initial_state)
    mach.states.initial.add(initial_state)
    # fix an ordering for keys
    # because tuple(dict.iteritems()) is not safe:
    # https://docs.python.org/2/library/stdtypes.html#dict.items
    try:
        u = next(iter(A))
        keys = A.node[u]['state'].keys()
    except Exception:
        logger.warn('strategy has no states.')
    # to store tuples of dict values for fast search
    isinit = spec.compile_init(no_str=True)
    # Mealy reaction to initial env input
    init_valuations = set()
    tmp = dict()
    for u, d in A.nodes_iter(data=True):
        var_values = d['state']
        vals = tuple(var_values[k] for k in keys)
        # already an initial valuation ?
        if vals in init_valuations:
            continue
        # add edge: Sinit -> u ?
        tmp.update(var_values)
        if eval(isinit, tmp):
            label = _int2str(var_values, str_vars)
            mach.transitions.add(initial_state, u, **label)
            # remember variable values to avoid
            # spurious non-determinism wrt the machine's memory
            #
            # in other words,
            # "state" omits the strategy's memory
            # hidden (existentially quantified)
            # so multiple nodes can be labeled with the same state
            #
            # non-uniqueness here would be equivalent to
            # multiple choices for initializing the hidden memory.
            init_valuations.add(vals)
            logger.debug('found initial state: {u}'.format(u=u))
        logger.debug('machine vertex: {u}, has var values: {v}'.format(
                     u=u, v=var_values))
    n = len(A)
    m = len(mach)
    assert m == n + 1, (n, m)
    if not mach.successors('Sinit'):
        raise Exception(
            'The machine obtained from the strategy '
            'does not have any initial states !\n'
            'The strategy is:\n'
            'vertices:' + pprint.pformat(A.nodes(data=True)) + 2 * '\n' +
            'edges:\n' + str(A.edges()) + 2 * '\n' +
            'and the machine:\n' + str(mach) + 2 * '\n' +
            'and the specification is:\n' + str(spec.pretty()) + 2 * '\n')
    return mach


def _int2str(label, str_vars):
    """Replace integers with string values for string variables.

    @param label: mapping from variable names, to integer (as strings)
    @type label: C{dict}

    @param str_vars: mapping that defines those variables that
        should be converted from integer to string variables.
        Each variable is mapped to a list of strings that
        comprise its range. This list defines how integer values
        correspond to string literals for that variable.
    @type str_vars: C{dict}

    @rtype: C{dict}
    """
    label = dict(label)
    label.update({k: str_vars[k][int(v)]
                 for k, v in label.iteritems()
                 if k in str_vars})
    return label


def mask_outputs(machine):
    """Erase outputs from each edge where they are zero."""
    for u, v, d in machine.edges_iter(data=True):
        for k in d:
            if k in machine.outputs and d[k] == 0:
                d.pop(k)


def determinize_machine_init(mach, init_out_values=None):
    """Return a determinized copy of C{mach} with given initial outputs.

    The transducers produced by synthesis can have multiple
    initial output valuations as possible reactions to a
    given input valuation.

    Possible reasons for this are:

      1. the system does not have full control over its initial state.
        For example the option "ALL_INIT" of C{gr1c}.

      2. the strategy returned by the solver has multiple
        vertices that satisfy the initial conditions.

    Case 1
    ======
    Requires an initial condition to be specified for
    each run of the transducer, because the transducer does
    not have full freedom to pick the initial output values.

    Note that solver options like "ALL_INIT"
    assume that the system has no control initially.
    Any output valuation that satisfies the initial
    conditions can occur.

    However, this may be too restrictive.
    The system may have control over the initial values of
    some outputs, but not others.

    For the outputs it can initially control,
    the non-determinism resulting from synthesis is redundancy
    and can be removed arbitrarily, as in Case 2.

    Case 2
    ======
    The function L{strategy2mealy} returns a transducer that
    for each initial input valuation,
    for each initial output valuation,
    reacts with a unique transition.

    But this can yield multile reactions to a single input,
    even for solver options like "ALL_ENV_EXIST_SYS_INIT" for C{gr1c}.
    The reason is that there can be multiple strategy vertices
    that satisfy the initial conditions, but the solver
    included them not because they are needed as initial reactions,
    but to be visited later by the strategy.

    These redundant initial reactions can be removed,
    and because the system has full control over their values,
    they can be removed in an arbitrary manner,
    keeping only a single reaction, for each input valuation.

    Algorithm
    =========
    Returns a deterministic transducer.
    This means that at each transducer vertex,
    for each input valuation,
    there is only a single reaction (output valuation) available.

    The non-determinism is resolved for the initial reaction
    by ensuring the outputs given in C{init_out_values}
    take those values.
    The remaining outputs are determinized arbitrarily.

    See also
    ========
    L{synthesize}, L{strategy2mealy}

    @param mach: possibly non-deterministic transducer,
        as produced, for example, by L{synthesize}.
    @type mach: L{MealyMachine}

    @param init_out_values: mapping from output ports that
        the system cannot control initially,
        to the initial values they take in this instance of the game.
    @type init_out_values: C{dict}

    @rtype: L{MealyMachine}
    """
    mach = copy.deepcopy(mach)
    if init_out_values is None:
        init_out_values = dict()
    '''determinize given outputs (uncontrolled)'''
    # restrict attention to given output ports only
    given_ports = tuple(k for k in mach.outputs if k in init_out_values)
    rm_edges = set()
    for i, j, key, d in mach.edges_iter(['Sinit'], data=True, keys=True):
        for k in given_ports:
            if d[k] != init_out_values[k]:
                rm_edges.add((i, j, key))
                break
    mach.remove_edges_from(rm_edges)
    '''determinize arbitrarily any remnant non-determinism'''
    # input valuations already seen
    # tuples of values used for efficiency (have __hash__)
    possible_inputs = set()
    # fix a key order
    inputs = tuple(k for k in mach.inputs)
    rm_edges = set()
    for i, j, key, d in mach.edges_iter(['Sinit'], data=True, keys=True):
        in_values = tuple(d[k] for k in inputs)
        # newly encountered input valuation ?
        if in_values not in possible_inputs:
            possible_inputs.add(in_values)
            continue
        else:
            rm_edges.add((i, j, key))
    mach.remove_edges_from(rm_edges)
    return mach



#!/usr/bin/env python
"""
Tests for transys.mathset (part of transys subpackage)
"""
from nose.tools import raises
from collections import Iterable

from tulip.transys.mathset import MathSet, SubSet, PowerSet, TypedDict
from tulip.transys.mathset import compare_lists, unique, contains_multiple
from tulip import transys as trs

def mathset_test():
    s = MathSet([1,2,[1,2] ] )
    q = MathSet()

    q.add(1)
    q |= [1,2]

    q |= ['a', 3, {3}]

    c = s | s

    d = s | q

    assert(c is not s)

    assert(q._set == {'a',1,2,3} )
    assert(q._list == [{3} ] )

    assert(c._set == {1,2} )
    assert(c._list == [[1,2] ] )

    assert(d._set == {1,2,3,'a'} )
    assert(d._list == [[1,2], {3} ] )

    assert(isinstance(s, Iterable) )
    assert(1 in s)
    assert(2 in s)
    assert([1, 2] in s)
    assert(5 not in q)

    assert(len(s) == 3)

    s.remove([1,2] )
    assert([1,2] not in s)
    assert(s._set == {1,2} )
    assert(s._list == [])
    assert(len(s) == 2)

    for item in s:
        print(item)

    """Mutable"""
    a = MathSet()
    b = MathSet([{'a':1} ] )

    a |= b

    assert(a._set == set() )
    assert(a._list == [{'a':1} ] )

class MathSet_operations_test:
    def setUp(self):
        self.x = MathSet(['a', 1, [1, 2], {'a', 'b', '8'} ] )
        self.y = MathSet(['b', -2, [3.5, 2.25], {'/', '_'} ] )
        self.small2_listnum = MathSet([[1,2], 0.2])
        self.small1_set = MathSet([{-1, 1}])

    def tearDown(self):
        self.x = None
        self.y = None

    def test_mul(self):
        assert self.small2_listnum * self.small1_set == \
            MathSet([([1,2], {-1, 1}), (0.2, {-1, 1})])
        assert self.small2_listnum * self.small2_listnum == \
            MathSet([([1,2], [1,2]), ([1,2], 0.2), (0.2, [1,2]), (0.2, 0.2)])

    def test_sub(self):
        assert self.small2_listnum - self.small2_listnum== MathSet()
        assert self.small1_set - self.small2_listnum == self.small1_set
        assert self.x - self.small2_listnum == \
            MathSet(['a', 1, {'a', 'b', '8'} ] )

    def test_isub(self):
        q = MathSet(self.small2_listnum)
        q -= self.small2_listnum
        assert q == MathSet()

        q = MathSet(self.small2_listnum)
        q -= MathSet()
        assert q == q

    def test_add_unhashable(self):
        self.small1_set.add([1,2])
        assert self.small1_set == MathSet([{-1, 1}, [1, 2]])

    def test_pop(self):
        # These tests could be improved by seeding random number
        # generation to ensure determinism.
        assert self.small1_set.pop() == {-1, 1}
        assert self.small1_set == MathSet()

        original = MathSet(self.small2_listnum)
        q = self.small2_listnum.pop()
        assert len(self.small2_listnum) + 1 == len(original)
        self.small2_listnum.add(q)
        assert self.small2_listnum == original

    def test_intersection(self):
        assert self.x.intersection(self.small2_listnum) == MathSet([[1, 2]])
        assert self.x.intersection(MathSet()) == MathSet()

    def test_intersects(self):
        assert self.x.intersects(self.small2_listnum)
        assert not self.small2_listnum.intersects(self.small1_set)


def unique_check(iterable, expected):
    print unique(iterable)
    assert unique(iterable) == expected

def unique_test():
    for (iterable, expected) in [(range(3), set([0, 1, 2])),
                                 ([], set([])),
                                 ([1, 1, -1], set([1, -1])),
                                 ([[1, 2], 3, 3], [[1, 2], 3]),
                                 ("Dessert!!", set("Desert!"))]:
        yield unique_check, iterable, expected

def contains_multiple_test():
    assert contains_multiple([1, 1])
    assert not contains_multiple(("cc",))
    assert contains_multiple("cc")


def test_tuple():
    s = MathSet((1,2))
    assert(s._set == {1,2})
    assert(s._list == [])

def subset_test():
    a = SubSet([1,2,3,4, {1:2} ] )
    print(a)

    a.add(1)
    a.add_from([1,2] )
    a |= [3,4]

    assert(a._set == {1,2,3,4} )

    a |= [{1:2} ]
    assert(a._list == [{1:2} ] )

    b = SubSet([1,'2'] )
    b.add('2')
    assert(b._set == {'2'} )
    assert(not bool(b._list) )
    assert(b._superset == [1,'2'] )

    superset = [1, 2]
    s = SubSet(superset)
    s |= [1, 2]
    print(s)
    assert(s._set == {1,2} )
    assert(not bool(s._list) )

    #s.add(3)

    return a

def powerset_test():
    s = [[1, 2], '3', {'a':1}, 1]

    p = PowerSet(s)

    s2 = ['3', {'a':1}, [1,2], 1]
    q = PowerSet()
    q.math_set = MathSet(s2)

    assert(p.math_set == MathSet(s) )
    assert(q.math_set == MathSet(s) )
    assert(isinstance(q.math_set, MathSet) )
    assert(p == q)

    q.math_set.add(6)

    # CAUTION: comparing p() and q() might yield False, due to ordering
    f = p +q
    assert(isinstance(f, PowerSet) )
    assert(f.math_set._set == {1, '3', 6} )
    assert(compare_lists(f.math_set._list, [[1, 2], {'a':1} ] ) )

    return s

class PowerSet_operations_test:
    def setUp(self):
        self.p = PowerSet({1, 2, 3})
        self.q_unhashable = PowerSet(MathSet([[1,2], ["a", "b"]]))
        self.singleton = PowerSet(set([1]))
        self.empty = PowerSet()

    def tearDown(self):
        self.p = None
        self.q_unhashable = None
        self.singleton = None

    def test_len(self):
        assert len(self.p) == 2**3
        assert len(self.q_unhashable) == 2**2
        assert len(self.singleton) == 2
        assert len(self.empty) == 1

    def test_call(self):
        p = [set(x) for x in self.p]
        assert len(p) == 2**3
        assert (set() in p)
        assert (set([1]) in p) and (set([2]) in p) and (set([3]) in p)
        assert (set([1,2]) in p) and (set([2,3]) in p) and (set([1,3]) in p)
        assert set([1,2,3]) in p
        assert set(self.singleton) == set([(), (1,)])
        assert set(self.empty) == set([()])

class TypedDict_test():
    def setUp(self):
        d = TypedDict()
        d.set_types({'animal':{'dog', 'cat'} })
        self.d = d

    def test_add_typed_key_value(self):
        d = self.d

        d['animal'] = 'dog'
        assert(d['animal'] == 'dog')

    @raises(ValueError)
    def test_add_typed_key_illegal_value(self):
        d = self.d

        d['animal'] = 'elephant'

    def test_add_untyped_key_value(self):
        d = self.d

        d['human'] = 'Bob'
        assert(d['human'] == 'Bob')



#!/usr/bin/env python
"""Installation script."""
import logging
import os
from setuptools import setup
# inline:
# import git


NAME = 'tulip'
VERSION_FILE = '{name}/_version.py'.format(name=NAME)
MAJOR = 1
MINOR = 3
MICRO = 0
VERSION = '{major}.{minor}.{micro}'.format(
    major=MAJOR, minor=MINOR, micro=MICRO)
VERSION_TEXT = (
    '# This file was generated from setup.py\n'
    "version = '{version}'\n")
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)
classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Topic :: Scientific/Engineering']
package_data = {
    'tulip.spec': ['parsetab.py']}


def package_jtlv():
    """Detect `jtlv`, or note its absence."""
    path = os.path.join('tulip', 'interfaces', 'jtlv_grgame.jar')
    if os.path.exists(path):
        print('Found optional JTLV-based solver.')
        package_data['tulip.interfaces'] = ['jtlv_grgame.jar']
    else:
        print('The jtlv synthesis tool was not found. '
              'Try extern/get-jtlv.sh to get it.\n'
              'It is an optional alternative to `omega`, '
              'the default GR(1) solver of TuLiP.')


def git_version(version):
    """Return version with local version identifier."""
    import git
    repo = git.Repo('.git')
    repo.git.status()
    sha = repo.head.commit.hexsha
    if repo.is_dirty():
        return '{v}.dev0+{sha}.dirty'.format(
            v=version, sha=sha)
    # commit is clean
    # is it release of `version` ?
    try:
        tag = repo.git.describe(
            match='v[0-9]*', exact_match=True,
            tags=True, dirty=True)
    except git.GitCommandError:
        return '{v}.dev0+{sha}'.format(
            v=version, sha=sha)
    assert tag == 'v' + version, (tag, version)
    return version


def run_setup():
    """Build parser, get version from `git`, install."""
    # Build PLY table, to be installed as tulip package data
    try:
        import tulip.spec.lexyacc
        tabmodule = tulip.spec.lexyacc.TABMODULE.split('.')[-1]
        outputdir = 'tulip/spec'
        parser = tulip.spec.lexyacc.Parser()
        parser.build(tabmodule, outputdir=outputdir,
                     write_tables=True,
                     debug=True, debuglog=logger)
        plytable_build_failed = False
    except Exception as e:
        print('Failed to build PLY tables: {e}'.format(e=e))
        plytable_build_failed = True
    # version
    try:
        version = git_version(VERSION)
    except AssertionError:
        raise
    except Exception:
        print('No git info: Assume release.')
        version = VERSION
    s = VERSION_TEXT.format(version=version)
    with open(VERSION_FILE, 'w') as f:
        f.write(s)
    # setup
    package_jtlv()
    setup(
        name=NAME,
        version=version,
        description='Temporal Logic Planning (TuLiP) Toolbox',
        author='Caltech Control and Dynamical Systems',
        author_email='tulip@tulip-control.org',
        url='http://tulip-control.org',
        bugtrack_url=('http://github.com/tulip-control/'
                      'tulip-control/issues'),
        license='BSD',
        classifiers=classifiers,
        install_requires=[
            'networkx >= 1.8, <= 1.10',
            'numpy >= 1.7',
            'omega >= 0.0.9, < 0.1.0',
            'ply >= 3.4',
            'polytope >= 0.1.2',
            'pydot >= 1.2.0',
            'scipy'],
        tests_require=[
            'nose',
            'matplotlib',
            'mock'],
        packages=[
            'tulip', 'tulip.transys', 'tulip.transys.export',
            'tulip.abstract', 'tulip.spec',
            'tulip.interfaces'],
        package_dir={'tulip': 'tulip'},
        package_data=package_data)
    # ply failed ?
    if plytable_build_failed:
        print('!' * 65 +
              '    Failed to build PLY table.  ' +
              'Please run setup.py again.' +
              '!' * 65)


if __name__ == '__main__':
    run_setup()



# Copyright (c) 2013 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Transition System module usage small examples
"""

import tulip.transys as trs

hl = 60*'='
save_fig = False

def fts_minimal_example():
    """Small example, for more see the maximal example."""

    fts = trs.FTS()
    fts.states.add_from(['s0', 's1'] )
    fts.states.initial.add('s0')

    fts.atomic_propositions.add_from({'green', 'not_green'})
    fts.states.add('s0', ap={'not_green'})
    fts.states.add('s1', ap={'green'})

    fts.transitions.add('s0', 's1')
    fts.transitions.add('s1', 's0')

    if not fts.plot() and save_fig:
        fts.save('small_fts.png')

    return fts

def ofts_minimal_example():
    """Open FTS demo."""
    msg = hl +'\nOpen FTS\n' +hl
    print(msg)

    ofts = trs.FiniteTransitionSystem()

    ofts.states.add_from(['s1', 's2', 's3'] )
    ofts.states.initial.add('s1')

    ofts.atomic_propositions |= ['p']
    ofts.states.add('s1', ap={'p'})
    ofts.states.add('s2', ap=set() )
    ofts.states.add('s3', ap={'p'})

    ofts.transitions.add('s1', 's2') # unlabeled

    ofts.sys_actions.add('try')
    ofts.sys_actions.add_from({'start', 'stop'} )
    ofts.env_actions.add_from({'block', 'wait'} )

    print(ofts.sys_actions)
    print(ofts.env_actions)

    # remove unlabeled edge s1->s2
    ofts.transitions.remove('s1', 's2')

    ofts.transitions.add(
        's1', 's2',
        sys_actions='try', env_actions='block'
    )
    ofts.transitions.add(
        's2', 's3',
        sys_actions='start', env_actions='wait'
    )
    ofts.transitions.add(
        's3', 's2',
        sys_actions='stop', env_actions='block'
    )

    print('The Open TS now looks like:')
    print(ofts.transitions() )

    ofts.atomic_propositions |= {'home', 'lot', 'p1'}

    print(ofts)

    path = './test_ofts'
    pdf_fname = path +'.pdf'

    if not ofts.plot() and save_fig:
        ofts.save(pdf_fname)

    return ofts

def ba_minimal_example():
    """Small example.

    ![]<>green  = <>[]!green

    ref
    ---
    Example 4.64, p.202 [Baier]

    note
    ----
    q2 state is a bit redundant, just let the automaton die.
    """

    msg = hl +'\nBuchi Automaton (small example):    '
    msg += 'Example 4.64, p.202 [Baier]\n' +hl
    print(msg)

    ba = trs.BuchiAutomaton(atomic_proposition_based=True)
    ba.states.add_from({'q0', 'q1', 'q2'})
    ba.states.initial.add('q0')
    ba.states.accepting.add('q1')

    ba.alphabet.math_set |= [True, 'green', 'not_green']

    ba.transitions.add('q0', 'q0', letter={True})
    ba.transitions.add('q0', 'q1', letter={'not_green'})
    ba.transitions.add('q1', 'q1', letter={'not_green'})
    ba.transitions.add('q1', 'q2', letter={'green'})
    ba.transitions.add('q2', 'q2', letter={True})

    if not ba.plot() and save_fig:
        ba.save('small_ba.png')

    return ba

def merge_example():
    """Merge two small FT Systems.
    """
    n = 4
    L = n*['p']
    ts1 = trs.line_labeled_with(L, n-1)

    ts1.transitions.remove('s3', 's4')
    ts1.transitions.remove('s5', 's6')

    ts1.sys_actions |= ['step', 'jump']
    ts1.transitions.add('s3', 's4', sys_actions='step')
    ts1.transitions.add('s5', 's6', sys_actions='jump')

    ts1.plot()

    L = n*['p']
    ts2 = trs.cycle_labeled_with(L)
    ts2.states.add('s3', ap=set())

    ts2.transitions.remove('s0', 's1')
    ts2.transitions.remove('s1', 's2')

    ts2.sys_actions |= ['up', 'down']
    ts2.transitions.add('s0', 's1', sys_actions='up')
    ts2.transitions.add('s1', 's2', sys_actions='down')

    ts2.plot()

    return ts2

if __name__ == '__main__':
    print('Intended to be run within IPython.\n'
          +'If no plots appear, change save_fig = True, '
          +'to save them to files instead.')

    fts = fts_minimal_example()
    ofts = ofts_minimal_example()
    ba = ba_minimal_example()

    merger_ts = merge_example()



"""test hybrid construction"""
import logging
logging.basicConfig(level=logging.INFO)

import time
import numpy as np

import matplotlib as mpl
mpl.use('Agg')

from tulip import abstract, hybrid
from polytope import box2poly

input_bound = 0.4
uncertainty = 0.05

cont_state_space = box2poly([[0., 3.], [0., 2.]])

cont_props = {}
cont_props['home'] = box2poly([[0., 1.], [0., 1.]])
cont_props['lot'] = box2poly([[2., 3.], [1., 2.]])

sys_dyn = dict()

allh = [0.5, 1.1, 1.5]

modes = []
modes.append(('normal', 'fly'))
modes.append(('refuel', 'fly'))
modes.append(('emergency', 'fly'))

"""First PWA mode"""
def subsys0(h):
    A = np.array([[1.1052, 0.], [ 0., 1.1052]])
    B = np.array([[1.1052, 0.], [ 0., 1.1052]])
    E = np.array([[1,0], [0,1]])

    U = box2poly([[-1., 1.], [-1., 1.]])
    U.scale(input_bound)

    W = box2poly([[-1., 1.], [-1., 1.]])
    W.scale(uncertainty)

    dom = box2poly([[0., 3.], [h, 2.]])

    sys_dyn = hybrid.LtiSysDyn(A, B, E, None, U, W, dom)

    return sys_dyn

def subsys1(h):
    A = np.array([[0.9948, 0.], [0., 1.1052]])
    B = np.array([[-1.1052, 0.], [0., 1.1052]])
    E = np.array([[1, 0], [0, 1]])

    U = box2poly([[-1., 1.], [-1., 1.]])
    U.scale(input_bound)

    W = box2poly([[-1., 1.], [-1., 1.]])
    W.scale(uncertainty)

    dom = box2poly([[0., 3.], [0., h]])

    sys_dyn = hybrid.LtiSysDyn(A, B, E, None, U, W, dom)

    return sys_dyn

for mode, h in zip(modes, allh):
    subsystems = [subsys0(h), subsys1(h)]
    sys_dyn[mode] = hybrid.PwaSysDyn(subsystems, cont_state_space)

"""Switched Dynamics"""

# collect env, sys_modes
env_modes, sys_modes = zip(*modes)
msg = 'Found:\n'
msg += '\t Environment modes: ' + str(env_modes)
msg += '\t System modes: ' + str(sys_modes)

switched_dynamics = hybrid.SwitchedSysDyn(
    disc_domain_size=(len(env_modes), len(sys_modes)),
    dynamics=sys_dyn,
    env_labels=env_modes,
    disc_sys_labels=sys_modes,
    cts_ss=cont_state_space
)

print(switched_dynamics)

ppp = abstract.prop2part(cont_state_space, cont_props)
ppp, new2old = abstract.part2convex(ppp)

"""Discretize to establish transitions"""
start = time.time()

N = 8
trans_len=1

disc_params = {}
for mode in modes:
    disc_params[mode] = {'N':N, 'trans_length':trans_len}

swab = abstract.multiproc_discretize_switched(
    ppp, switched_dynamics, disc_params,
    plot=True, show_ts=True
)
print(swab)
axs = swab.plot(show_ts=True)
for i, ax in enumerate(axs):
    ax.figure.savefig('swab_' + str(i) + '.pdf')

#ax = sys_ts.ts.plot()

elapsed = (time.time() - start)
print('Discretization lasted: ' + str(elapsed))



from tulip import hybrid
import polytope
import scipy.io

"""
Contains functions that will read a .mat file exported by the MATLAB function
mpt2python and import it to either a PwaSysDyn or a LtiSysDyn.

Stephanie Tsuei, June 2014
"""


def load(filename):
    data = scipy.io.loadmat(filename)
    islti = bool(data['islti'][0][0])
    ispwa = bool(data['ispwa'][0][0])

    if islti:
        sys = load_lti(data['A'], data['B'], data['domainA'],
                       data['domainB'], data['UsetA'], data['UsetB'])

    elif ispwa:
        nlti = len(data['A'][0])
        lti_systems = []

        for i in xrange(nlti):
            A = data['A'][0][i]
            B = data['B'][0][i]
            K = data['K'][0][i]
            domainA = data['domainA'][0][i]
            domainB = data['domainB'][0][i]
            UsetA = data['UsetA'][0][i]
            UsetB = data['UsetB'][0][i]

            ltisys = load_lti(A, B, K, domainA, domainB, UsetA, UsetB)
            lti_systems.append(ltisys)

        cts_ss = polytope.Polytope(data['ctsA'], data['ctsB'])
        sys = hybrid.PwaSysDyn(list_subsys=lti_systems, domain=cts_ss)

    return sys



def load_lti(A, B, K, domainA, domainB, UsetA, UsetB):
    domain = polytope.Polytope(domainA, domainB)
    Uset = polytope.Polytope(UsetA, UsetB)

    lti = hybrid.LtiSysDyn(A=A, B=B, K=K, domain=domain, Uset=Uset)
    return lti



# Copyright (c) 2014, 2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Interface to ltl2ba"""
import logging
logger = logging.getLogger(__name__)
import subprocess
import networkx as nx
import ply.lex
import ply.yacc


TABMODULE = 'ltl2ba_parsetab'


class Lexer(object):
    """Token rules to build lexer for ltl2ba output."""

    reserved = {
        'goto': 'GOTO',
        'if': 'IF',
        'fi': 'FI',
        'never': 'NEVER',
        'skip': 'SKIP'}

    tokens = (
        'TRUE', 'FALSE',
        'NUMBER',
        'NOT', 'AND', 'OR', 'XOR', 'IMP', 'BIMP',
        'EQUALS', 'NEQUALS', 'LT', 'LE', 'GT', 'GE',
        'PLUS', 'MINUS', 'TIMES', 'DIV',
        'LPAREN', 'RPAREN', 'LBRACE', 'RBRACE',
        'COLON', 'COLON2', 'SEMI',
        'COMMENT', 'NAME') + tuple(reserved.values())

    # Tokens
    t_TRUE = 'TRUE|True|true'
    t_FALSE = 'FALSE|False|false'

    t_COMMENT = '/\*.*\*/'
    t_NOT = r'\!'
    t_AND = r'\&\&|\&'
    t_OR = r'\|\||\|'
    t_XOR = r'\^'

    t_EQUALS = r'\=|\=\='
    t_NEQUALS = r'\!\='
    t_LT = r'\<'
    t_LE = r'\<\='
    t_GT = r'>\='
    t_GE = r'>'

    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_LBRACE = r'\{'
    t_RBRACE = r'\}'
    t_SEMI = r';'
    t_COLON2 = r'::'
    t_COLON = r':'

    t_NUMBER = r'\d+'

    t_IMP = r'->'

    # Ignored characters
    t_ignore = ' \t'

    def __init__(self):
        self.lexer = ply.lex.lex(module=self, debug=False)

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")

    def t_error(self, t):
        logger.warn("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    def t_name(self, t):
        r'[A-Za-z_][a-zA-Z0-9_]*'
        t.type = self.reserved.get(t.value, 'NAME')
        return t


class Parser(object):
    """Production rules to build parser for ltl2ba output."""

    precedence = (
        # ('right', 'UNTIL', 'RELEASE'),
        ('right', 'BIMP'),
        ('right', 'IMP'),
        ('left', 'XOR'),
        ('left', 'OR'),
        ('left', 'AND'),
        # ('right', 'ALWAYS', 'EVENTUALLY'),
        # ('right', 'NEXT'),
        ('right', 'NOT'),
        # ('left', 'PRIME'),
        ('nonassoc', 'EQUALS', 'NEQUALS', 'LT', 'LE', 'GT', 'GE'),
        ('nonassoc', 'TIMES', 'DIV'),
        ('nonassoc', 'PLUS', 'MINUS'),
        ('nonassoc', 'TRUE', 'FALSE'))

    def __init__(self):
        """Build lexer and parser."""
        self.lexer = Lexer()
        self.tokens = self.lexer.tokens
        self.parser = ply.yacc.yacc(module=self, tabmodule=TABMODULE,
                                    write_tables=True, debug=False)
        self.g = None
        self.initial_nodes = None
        self.accepting_nodes = None
        self.symbols = None

    def parse(self, ltl2ba_output):
        """Return a Buchi automaton from parsing C{ltl2ba_output}.

        @return: Buchi automaton as a 3-`tuple` containing:
          - `dict` mapping symbols to types (all `"boolean"`)
          - `networkx.MultiDiGraph`, each edge labeled with a
            `"guard"` that is a Boolean formula as `str`
          - `set` of initial nodes
          - `set` of accepting nodes
        """
        self.g = nx.MultiDiGraph()
        # flat is better than nested
        self.initial_nodes = set()
        self.accepting_nodes = set()
        self.symbols = dict()
        self.parser.parse(ltl2ba_output, lexer=self.lexer.lexer)
        return (self.symbols, self.g,
                self.initial_nodes, self.accepting_nodes)

    def p_claim(self, p):
        """claim : NEVER LBRACE COMMENT clauses RBRACE"""
        p[0] = p[4]

    def p_clauses(self, p):
        """clauses : clauses clause"""
        p[0] = p[1] + [p[2]]

    def p_clauses_end(self, p):
        """clauses : clause"""
        p[0] = [p[1]]

    def p_clause(self, p):
        """clause : if_clause"""
        p[0] = p[1]

    def p_clause_labeled(self, p):
        """clause : state COLON if_clause"""
        u = p[1]
        if_clause = p[3]
        for guard, v in if_clause:
            self.g.add_edge(u, v, guard=guard)
        p[0] = (u, if_clause)

    def p_if_clause(self, p):
        """if_clause : IF cases FI SEMI"""
        p[0] = p[2]

    def p_cases(self, p):
        """cases : cases case"""
        p[0] = p[1] + [p[2]]

    def p_cases_end(self, p):
        """cases : case"""
        p[0] = [p[1]]

    def p_case(self, p):
        """case : COLON2 expr IMP goto"""
        p[0] = (p[2], p[4])

    def p_expr_paren(self, p):
        """expr : LPAREN expr RPAREN"""
        p[0] = '({expr})'.format(expr=p[2])

    def p_and(self, p):
        """expr : expr AND expr"""
        p[0] = '({x} and {y})'.format(x=p[1], y=p[3])

    def p_or(self, p):
        """expr : expr OR expr"""
        p[0] = '({x} or {y})'.format(x=p[1], y=p[3])

    def p_not(self, p):
        """expr : NOT expr"""
        p[0] = '(not {expr})'.format(expr=p[2])

    def p_number(self, p):
        """expr : NUMBER"""
        p[0] = int(p[1])

    def p_expr_name(self, p):
        """expr : NAME"""
        self.symbols[p[1]] = 'boolean'
        p[0] = p[1]

    def p_goto(self, p):
        """goto : GOTO state"""
        p[0] = p[2]

    def p_state(self, p):
        """state : NAME"""
        state = p[1]
        self.g.add_node(state)
        if 'init' in state:
            self.initial_nodes.add(state)
        elif 'accept' in state:
            self.accepting_nodes.add(state)
        p[0] = p[1]

    def p_empty(self, p):
        """empty :"""

    def p_error(self, p):
        logger.error('Syntax error at ' + p.value)


def call_ltl2ba(formula, prefix=''):
    """Load a Buchi Automaton from a Never Claim.

    TODO
    ====
    Make sure guard semantics properly accounted for:
    'prop' | '!prop' | '1' and skip

    Depends
    =======
    ltl2ba: http://www.lsv.ens-cachan.fr/~gastin/ltl2ba/

    @param formula: LTL formula for input to ltl2ba
    @type formula: str

    @return: Buchi Automaton
    @rtype: tulip.transys.BA
    """
    try:
        subprocess.call(['ltl2ba', '-h'], stdout=subprocess.PIPE)
    except OSError:
        raise Exception('cannot find ltl2ba on path')
    p = subprocess.Popen(
        [prefix + 'ltl2ba', '-f', '"{f}"'.format(f=formula)],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)
    p.wait()
    ltl2ba_output = p.stdout.read()
    logger.info('ltl2ba output:\n\n{s}\n'.format(s=ltl2ba_output))
    if p.returncode != 0:
        raise Exception('Error when converting LTL to Buchi.')
    return ltl2ba_output


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    logger.setLevel(level=logging.DEBUG)
    parser = Parser()
    f = '[] !s && <> (a || c) && <> (b U c)'
    out = call_ltl2ba(f)
    symbols, g, initial, accepting = parser.parse(out)
    g.save('ba.pdf')



"""
Aerial refueling double tank balancing example.

The system consists of two fuel tanks, T1 and T2.
Fuel is drawn at a constant rate from T2.
There are two dynamic modes:

    - 'normal operation' and
    - 'air refuel mode'.

In 'air refuel mode' fuel is added
to T1 at a constant rate.

The objective is to keep the fuel levels in
the tanks close to each other.

CODE KEY:
- ## : beginning of major step; e.g., dynamics definition

For a smaller example, consider adjusting the script to use instead
the following parameters:
tank_capacity = 5
input_ub = 10
max_vol_diff = 1
max_refuel_level = 3
init_lower = 1
init_upper = 3

reference
=========
    Nilsson P.; Ozay N.; Topcu U.; Murray R.M.
     Temporal Logic Control of Switched Affine Systems
     with an Application in Fuel Balancing
    2012 American Control Conference
"""
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

logging.getLogger('tulip').setLevel(logging.ERROR)
logging.getLogger('tulip.synth').setLevel(logging.ERROR)

log = logging.getLogger('multiprocessing')
#log.setLevel(logging.ERROR)

import os
import pickle
import numpy as np
#from scipy import io as sio
#import matplotlib
import matplotlib as mpl
mpl.use('Agg')

from tulip import hybrid, abstract, spec, synth
import polytope as pc
from tulip.abstract.plot import plot_strategy
#from tulip.graphics import newax

## Problem variables
tank_capacity = 10      # Maximum tank capacity
fuel_consumption = 1    # Rate at which fuel is drawn from tank 2
refill_rate = 3         # Rate at which fuel is refilled in tank 1 during refuel mode
input_lb = 0            # Lower bound on fuel move rate from tank 1 to 2
input_ub = 3            # Upper bound on fuel move rate from tank 1 to 2
max_vol_diff = 2        # Enforce |v1-v2| to be lower than this value
fin_vol_diff = 1        # Enforce |v1-v2| to always eventually. be lower than this value
max_refuel_level = 8    # Above this level no refueling will take place
N = 1                   # Horizon length
disturbance = 0.0       # Absolute uncertainty in fuel consumption
init_lower = 6          # Lower bound for possible initial volumes
init_upper = 8          # Upper bound for possible initial volumes
fast = True             # Use smaller domain to increase speed

imgpath = './'
fontsize = 18

## State space and propositions
if fast:
    cont_ss = pc.Polytope(
        np.array([[1,0],
                  [-1,0],
                  [0,1],
                  [0,-1],
                  [1,-1],
                  [-1,1]]),
        np.array([tank_capacity, 1, tank_capacity,
                  1, 2*max_vol_diff, 2*max_vol_diff])
    )
else:
    cont_ss = pc.Polytope(
        np.array([[1,0],
                  [-1,0],
                  [0,1],
                  [0,-1]]),
        np.array([tank_capacity, 0, tank_capacity, 0])
    )

cont_props = {}
cont_props['no_refuel'] = pc.Polytope(
    np.array([[1,0],
              [-1,0],
              [0,1],
              [0,-1]]),
    np.array([tank_capacity, 0, tank_capacity, -max_refuel_level])
)
cont_props['vol_diff'] = pc.Polytope(
    np.array([[-1,0],
              [0,-1],
              [-1,1],
              [1,-1]]),
    np.array([0,0,max_vol_diff,max_vol_diff])
)
cont_props['vol_diff2'] = pc.Polytope(
    np.array([[-1,0],
              [0,-1],
              [-1,1],
              [1,-1]]),
    np.array([0,0,fin_vol_diff,fin_vol_diff])
)
cont_props['initial'] = pc.Polytope(
    np.array([[1,0],
             [-1,0],
             [0, 1],
             [0,-1]]),
    np.array([init_upper, -init_lower, init_upper, -init_lower])
)
cont_props['critical'] = pc.Polytope(
    np.array([[-1,0],
              [0,-1],
              [1,1]]),
    np.array([0, 0, 2*fuel_consumption])
)

## Dynamics
A = np.eye(2)
B = np.array([[-1],[1]])
E = np.array([[0],[1]])
K1 = np.array([[0.],[-fuel_consumption]])
K2 = np.array([[refill_rate],[-fuel_consumption]])
U1 = pc.Polytope(
    np.array([
        [1,0,0],
        [-1,0,0],
        [1,-1,0]
    ]),
    np.array([input_ub, -input_lb, 0])
)
U2 = pc.Polytope(
    np.array([
        [1,0,0],
        [-1,0,0],
        [1,-1,0]
    ]),
    np.array([input_ub,-input_lb, refill_rate])
)
W = pc.Polytope(np.array([[1],[-1]]),
                np.array([disturbance, disturbance]))

# Normal operation dynamics
cont_dyn_normal = hybrid.LtiSysDyn(A, B, E, K1, U1, W, domain=cont_ss)

# Aerial refueling mode dynamics
cont_dyn_refuel = hybrid.LtiSysDyn(A, B, E, K2, U2, W, domain=cont_ss)

## Switched Dynamics
env_modes = ('normal', 'refuel')
sys_modes = ('fly',)

pwa_normal = hybrid.PwaSysDyn([cont_dyn_normal], domain=cont_ss)
pwa_refuel = hybrid.PwaSysDyn([cont_dyn_refuel], domain=cont_ss)

dynamics_dict = {
    ('normal', 'fly') : pwa_normal,
    ('refuel', 'fly') : pwa_refuel
}

switched_dynamics = hybrid.SwitchedSysDyn(
    cts_ss=cont_ss,
    disc_domain_size=(len(env_modes), len(sys_modes)),
    dynamics=dynamics_dict,
    env_labels=env_modes,
    disc_sys_labels=sys_modes
)

## Create convex proposition preserving partition
ppp = abstract.prop2part(cont_ss, cont_props)
ppp, new2old = abstract.part2convex(ppp)

ax = ppp.plot_props()
ax.figure.savefig(imgpath + 'cprops.pdf')

ax = ppp.plot()
ax.figure.savefig(imgpath + 'ppp.pdf')

## Discretize to establish transitions
if os.name == "posix":
    start = os.times()[2]
    logger.info('start time: ' + str(start))
else:
    logger.info('Timing currently only available for POSIX platforms (not Windows)')

disc_params = {}
disc_params[('normal', 'fly')] = {'N':N, 'trans_length':3}
disc_params[('refuel', 'fly')] = {'N':N, 'trans_length':3}

sys_ts = abstract.multiproc_discretize_switched(
    ppp, switched_dynamics, disc_params, plot=True
)

if os.name == "posix":
    end = os.times()[2]
    logger.info('end time: ' + str(end))
    elapsed = (end - start)
    logger.info('Discretization lasted: ' + str(elapsed))

## Save abstraction to save debugging time
fname = './abstract_switched.pickle'
#pickle.dump(sys_ts, open(fname, 'wb') )

#sys_ts = pickle.load(open(fname, 'r') )

## Specifications
env_vars = set()
sys_disc_vars = set()

env_init = {'env_actions = "normal"'}
#env_init |= {'initial'}

env_safe = {'no_refuel -> X(env_actions = "normal")',
            '(critical & (env_actions = "normal")) -> X(env_actions = "refuel")',
            '(!critical & env_actions = "normal") -> X(env_actions = "normal")',
            '(!no_refuel & env_actions = "refuel") -> X(env_actions = "refuel")'}
env_prog = {'env_actions = "refuel"'}

# relate switching actions to u_in (env_actions)
sys_init = {'initial'}
sys_safe = {'vol_diff'}
sys_prog = {'True'} #{'vol_diff2'}

specs = spec.GRSpec(env_vars, sys_disc_vars,
                    env_init, sys_init,
                    env_safe, sys_safe,
                    env_prog, sys_prog)
print(specs.pretty())

## Synthesis
print("Starting synthesis")
if os.name == "posix":
    start = os.times()[2]

specs.moore = False
specs.qinit = r'\A \E'
ctrl = synth.synthesize(
    'omega', specs, sys=sys_ts.ts, ignore_sys_init=True,
    #action_vars=('u_in', 'act')
)
if os.name == "posix":
    end = os.times()[2]
    elapsed = (end - start)
    logger.info('Synthesis lasted: ' + str(elapsed))

logger.info(ctrl)
ctrl.save(imgpath + 'double_tank.pdf')

ax = plot_strategy(sys_ts, ctrl)
ax.figure.savefig(imgpath + 'proj_mealy.pdf')

## Simulate
# num_it = 25
# init_state = {}
# init_state['u_in'] = 0

# destfile = 'rsdisturbance_example.gexf'
# states = grsim.grsim(
#     [aut], env_states=[init_state], num_it=num_it,
#     deterministic_env=False, graph_vis=False,
#     destfile=destfile
# )

# uin_arr = []
# cellid_arr = []
# for (autID, state) in states:
#     uin_arr.append(state.state['u_in'])
#     cellid_arr.append(state.state['cellID'])
# uin_arr = np.array(uin_arr)
# cellid_arr = np.array(cellid_arr)

# rc, x = pc.cheby_ball(new_part.regions[cellid_arr[0]])
# x = x.flatten()
# x_arr = x.copy()
# u_arr = np.zeros(1)
# for i in range(1, len(cellid_arr)):
#     if uin_arr[i-1] == 0:
#         u = abstract.get_input(
#             x, cont_dyn_normal, new_part,
#             cellid_arr[i-1], cellid_arr[i], 1, mid_weight=10
#         )
#         x = np.dot(cont_dyn_normal.A,x).flatten() + \
#             np.dot(cont_dyn_normal.B,u).flatten() + \
#             cont_dyn_normal.K.flatten()
#     else:
#         u = abstract.get_input(
#             x, cont_dyn_refuel, new_part, cellid_arr[i-1],
#             cellid_arr[i], 1, Q=[], mid_weight=10
#         )
#         x = np.dot(cont_dyn_refuel.A,x).flatten() + \
#             np.dot(cont_dyn_refuel.B,u).flatten() + \
#             cont_dyn_refuel.K.flatten()

#     u_arr = np.hstack([u_arr, u.flatten()])
#     x_arr = np.vstack([x_arr, x])

# data = {}
# data['x'] = x_arr
# data['u'] = u_arr[range(1,u_arr.shape[0]), :]
# data['u_in'] = uin_arr

# sio.savemat("matlabdata", data)

# ax = plot_partition(new_part, plot_numbers=False, show=False)
# for tick in ax.xaxis.get_major_ticks():
#     tick.label1.set_fontsize(fontsize)
# for tick in ax.yaxis.get_major_ticks():
#     tick.label1.set_fontsize(fontsize)
# plt.xlabel('$v_1$', fontsize=fontsize+6)
# plt.ylabel('$v_2$', fontsize=fontsize+6)

# # Plot transitions
# arr_size=0.2
# for i in range(1,x_arr.shape[0]):
#     x = x_arr[i-1,0]
#     y = x_arr[i-1,1]
#     dx = x_arr[i,0] - x
#     dy = x_arr[i,1] - y
#     arr = matplotlib.patches.Arrow(
#         float(x), float(y), float(dx),
#         float(dy), width=arr_size
#     )
#     ax.add_patch(arr)

# ax.plot(x_arr[0,0], x_arr[0,1], 'og')
# ax.plot(x_arr[-1,0], x_arr[-1,1], 'or')

# plt.savefig(imgpath + 'simulation.pdf')



# Copyright (c) 2011-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""Abstract syntax tree classes for LTL.

Syntax taken originally roughly from:
http://spot.lip6.fr/wiki/LtlSyntax
"""
import logging
logger = logging.getLogger(__name__)
from abc import ABCMeta, abstractmethod


# prototype for flattening to a "canonical" string
OPMAP = {
    'False': 'False', 'True': 'True',
    '!': '!',
    '|': '|', '&': '&', '->': '->', '<->': '<->', '^': '^', 'ite': 'ite',
    'X': 'X', 'G': 'G', 'F': 'F',
    'U': 'U', 'W': 'W', 'V': 'V',
    '<': '<', '<=': '<=', '=': '=', '>=': '>=', '>': '>', '!=': '!=',
    '+': '+', '-': '-', '<<>>': '<<>>'  # arithmetic
}

# this mapping is based on SPIN documentation:
#   http://spinroot.com/spin/Man/ltl.html
FULL_OPERATOR_NAMES = {
    'next': 'X',
    'always': '[]',
    'eventually': '<>',
    'until': 'U',
    'stronguntil': 'U',
    'weakuntil': 'W',
    'unless': 'W',  # see Baier - Katoen
    'release': 'V',
    'implies': '->',
    'equivalent': '<->',
    'not': '!',
    'and': '&&',
    'or': '||',
}


def make_nodes(opmap=None):
    """Return class with attributes the AST node classes.

    The tree is defined recursively,
    not with a graph data structure.
    L{Tree} is a graph data structure for that purpose.
    """
    if opmap is None:
        opmap = OPMAP

    class Node(object):
        """Base class for AST nodes."""

        # Caution
        # =======
        # Do **NOT** implement C{__hash__}, because you
        # will unintendently identify different AST nodes !
        # Only leaf nodes can be safely identified.
        #
        # The default for user-defined classes is
        # C{__hash__ == _}
        __metaclass__ = ABCMeta
        opmap = None

        @abstractmethod
        def __init__(self):
            pass

        @abstractmethod
        def __repr__(self):
            pass

        @abstractmethod
        def flatten(self):
            pass

    Node.opmap = opmap

    # Do not confuse "term" with the abbreviation of "terminal".
    # A "term" in FOL can comprise of terminals,
    # for example a function together with parentheses and its args.
    class Terminal(Node):
        """Terminal symbols of grammar.

        Include:

          - 0-ary function constants (numbers, strings)
          - 0-ary function variables (integer or string variable)
          - 0-ary connectives (Boolean constants)
          - 0-ary predicate constants
          - 0-ary predicate variables
        """

        def __init__(self, value):
            if not isinstance(value, basestring):
                raise TypeError(
                    'value must be a string, got: {v}'.format(
                        v=value))
            self.type = 'terminal'
            self.value = value

        def __repr__(self):
            return '{t}({v})'.format(t=type(self).__name__,
                                     v=repr(self.value))

        def __str__(self, *arg, **kw):
            # *arg accommodates "depth" arg of Operator.__str__
            return self.value

        def __len__(self):
            """Return the number of operators and terminals.

            Note that this definition differs from the
            theoretical definition that a formula's length
            is the number of operators it contains.
            """
            return 1

        def __eq__(self, other):
            return (isinstance(other, type(self)) and
                    self.value == other.value)

        def flatten(self, *arg, **kw):
            return self.value

    class Operator(Node):
        """Takes a non-zero number of operands and returns a result.

        Cases:

          - function (arithmetic):
              maps (terms)^n to terms
          - predicate (relational operator):
              maps (terms)^n to atomic formulas
          - connective (logical operator):
              maps (wff)^n to wff
        """

        def __init__(self, operator, *operands):
            try:
                operator + 'a'
            except TypeError:
                raise TypeError(
                    'operator must be string, got: {op}'.format(
                        op=operator))
            self.type = 'operator'
            self.operator = operator
            self.operands = list(operands)

        # ''.join would be faster, but __repr__ is for debugging,
        # not for flattening, so readability takes precedence
        def __repr__(self):
            return '{t}({op}, {xyz})'.format(
                t=type(self).__name__,
                op=repr(self.operator),
                xyz=', '.join(repr(x) for x in self.operands))

        # more readable recursive counterpart of __repr__
        # depth allows limiting recursion to see a shallower view
        def __str__(self, depth=None):
            if depth is not None:
                depth = depth - 1
            if depth == 0:
                return '...'
            return '({op} {xyz})'.format(
                op=self.operator,
                xyz=' '.join(x.__str__(depth=depth)
                             for x in self.operands))

        def __len__(self):
            return 1 + sum(len(x) for x in self.operands)

        def flatten(self, *arg, **kw):
            return ' '.join([
                '(',
                self.opmap[self.operator],
                ', '.join(x.flatten(*arg, **kw) for x in self.operands),
                ')'])

    # Distinguish operators by arity
    class Unary(Operator):
        pass

    class Binary(Operator):
        def flatten(self, *arg, **kw):
            """Infix flattener for consistency with parser.

            Override it if you want prefix or postfix.
            """
            return ' '.join([
                '(',
                self.operands[0].flatten(*arg, **kw),
                self.opmap[self.operator],
                self.operands[1].flatten(*arg, **kw),
                ')'])

    class Nodes(object):
        """AST nodes for a generic grammar."""

    nodes = Nodes()
    nodes.Node = Node
    nodes.Terminal = Terminal
    nodes.Operator = Operator
    nodes.Unary = Unary
    nodes.Binary = Binary
    return nodes


def make_fol_nodes(opmap=None):
    """AST classes for fragment of first-order logic."""
    nodes = make_nodes(opmap)

    class Var(nodes.Terminal):
        """A 0-ary variable.

        Two cases:

          - 0-ary function variable (integer or string variable)
          - 0-ary propositional variable (atomic proposition)
        """

        def __init__(self, value):
            super(Var, self).__init__(value)
            self.type = 'var'

    class Bool(nodes.Terminal):
        """A 0-ary connective."""

        def __init__(self, value):
            if not isinstance(value, basestring):
                raise TypeError(
                    'value must be string, got: {v}'.format(v=value))
            if value.lower() not in {'true', 'false'}:
                raise TypeError(
                    'value must be "true" or "false" '
                    '(case insensitive), got: {v}'.format(v=value))
            self.value = 'True' if (value.lower() == 'true') else 'False'
            self.type = 'bool'

        def flatten(self, *arg, **kw):
            return self.opmap[self.value]

    class Num(nodes.Terminal):
        """A 0-ary function."""
        # self.value is str,
        # use int(self.value) if you need to

        def __init__(self, value):
            super(Num, self).__init__(value)
            self.type = 'num'

    class Str(nodes.Terminal):
        """A 0-ary function."""
        # parser ensures that value has no quotes

        def __init__(self, value):
            super(Str, self).__init__(value)
            self.type = 'str'

    class Comparator(nodes.Binary):
        """Binary relational operator (2-ary predicate)."""

    class Arithmetic(nodes.Binary):
        """Binary function.

        Maps terms to terms.
        """

    nodes.Var = Var
    nodes.Bool = Bool
    nodes.Num = Num
    nodes.Str = Str
    nodes.Comparator = Comparator
    nodes.Arithmetic = Arithmetic
    return nodes


nodes = make_fol_nodes()



# Copyright (c) 2013-2015 by California Institute of Technology
# and 2014 The Regents of the University of Michigan
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder(s) nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Base classes for labeled directed graphs"""
from __future__ import absolute_import
import logging
import os
import copy
from pprint import pformat
from collections import Iterable
import warnings
import networkx as nx
from tulip.transys.mathset import SubSet, TypedDict
# inline imports:
#
# from tulip.transys.export import graph2dot
# from tulip.transys.export import save_d3
# from tulip.transys.export import graph2dot


logger = logging.getLogger(__name__)


def label_is_desired(attr_dict, desired_dict):
    """Return True if all labels match.

    Supports symbolic evaluation, if label type is callable.
    """
    if not isinstance(attr_dict, TypedDict):
        raise Exception('attr_dict must be TypedDict' +
                        ', instead: ' + str(type(attr_dict)))
    if attr_dict == desired_dict:
        return True
    # different keys ?
    mismatched_keys = set(attr_dict).symmetric_difference(desired_dict)
    if mismatched_keys:
        return False
    # any labels have symbolic semantics ?
    label_def = attr_dict.allowed_values
    for type_name, value in attr_dict.iteritems():
        logger.debug('Checking label type:\n\t' + str(type_name))
        type_def = label_def[type_name]
        desired_value = desired_dict[type_name]
        if hasattr(type_def, '__call__'):
            logger.debug('Found label semantics:\n\t' + str(type_def))
            # value = guard
            if not type_def(value, desired_value):
                return False
            else:
                continue
        # no guard semantics given,
        # then by convention: guard is singleton {cur_val},
        if not value == desired_value:
            test_common_bug(value, desired_value)
            return False
    return True


def test_common_bug(value, desired_value):
    logger.debug('Label value:\n\t' + str(value))
    logger.debug('Desired value:\n\t' + str(desired_value))
    if (
        isinstance(value, (set, list)) and
        isinstance(desired_value, (set, list)) and
        value.__class__ != desired_value.__class__
    ):
        msg = (
            'Set SubLabel:\n\t' + str(value) +
            'compared to list SubLabel:\n\t' + str(desired_value) +
            'Did you mix sets & lists when setting AP labels ?')
        raise Exception(msg)


class States(object):
    """Methods to manage states and initial states."""

    def __init__(self, graph):
        """Initialize C{States}.

        @type graph: L{LabeledDiGraph}
        """
        self.graph = graph
        self.initial = []

    def __getitem__(self, state):
        return self.graph.node[state]

    def __call__(self, *args, **kwargs):
        """Return list of states.

        For more details see L{LabeledDiGraph.nodes}
        """
        return self.graph.nodes(*args, **kwargs)

    def __str__(self):
        return 'States:\n' + pformat(self(data=False))

    def __len__(self):
        """Total number of states."""
        return self.graph.number_of_nodes()

    def __iter__(self):
        return iter(self())

    def __ior__(self, new_states):
        # TODO carefully test this
        self.add_from(new_states)
        return self

    def __contains__(self, state):
        """Return True if state in states."""
        return state in self.graph

    @property
    def initial(self):
        """ Return L{SubSet} of initial states."""
        return self._initial

    @initial.setter
    def initial(self, states):
        s = SubSet(self)
        s |= states
        self._initial = s

    def _warn_if_state_exists(self, state):
        if state in self:
            logger.debug('State already exists.')

    def _single_state2singleton(self, state):
        """Convert to a singleton list, if argument is a single state.

        Otherwise return given argument.
        """
        if state in self:
            states = [state]
        else:
            states = state
        return states

    def add(self, new_state, attr_dict=None, check=True, **attr):
        """Wraps L{LabeledDiGraph.add_node},

        which wraps C{networkx.MultiDiGraph.add_node}.
        """
        self._warn_if_state_exists(new_state)
        logger.debug('Adding new id: ' + str(new_state))
        self.graph.add_node(new_state, attr_dict, check, **attr)

    def add_from(self, new_states, check=True, **attr):
        """Wraps L{LabeledDiGraph.add_nodes_from},

        which wraps C{networkx.MultiDiGraph.add_nodes_from}.
        """
        self.graph.add_nodes_from(new_states, check, **attr)

    def remove(self, state):
        """Remove C{state} from states (including initial).

        Wraps C{networkx.MultiDiGraph.remove_node}.
        """
        if state in self.initial:
            self.initial.remove(state)
        self.graph.remove_node(state)

    def remove_from(self, states):
        """Remove a list of states.

        Iterates C{States.remove} to imitate
        C{networkx.MultiDiGraph.remove_nodes_from},
        handling also initial states.
        """
        for state in states:
            self.remove(state)

    def post(self, states=None):
        """Direct successor set (1-hop) for given states.

        Edge labels are ignored.

        If multiple states provided,
        then union Post(s) for s in states provided.

        See Also
        ========
          - L{pre}
          - Def. 2.3, p.23 U{[BK08]
            <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}

        @param states:
          - None, so initial states returned
          - single state or
          - set of states or

        @rtype: set
        """
        if states is None:
            return set(self.initial)
        states = self._single_state2singleton(states)
        successors = set()
        for state in states:
            successors |= set(self.graph.successors(state))
        return successors

    def pre(self, states):
        """Return direct predecessors (1-hop) of given state.

        See Also
        ========
          - L{post}
          - Def. 2.3, p.23 U{[BK08]
            <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}

        @rtype: set
        """
        states = self._single_state2singleton(states)
        predecessors = set()
        for state in states:
            predecessors |= set(self.graph.predecessors(state))
        return predecessors

    def forward_reachable(self, state):
        """Return states reachable from given state.

        Iterated post(), a wrapper of networkx.descendants.
        """
        descendants = nx.descendants(self, state)
        return descendants

    def backward_reachable(self, state):
        """Return states from which the given state can be reached.

        A wrapper of networkx.ancestors.
        """
        ancestors = nx.ancestors(self, state)
        return ancestors

    def paint(self, state, color):
        """Color the given state.

        The state is filled with given color,
        rendered with dot when plotting and saving.

        @param state: valid system state

        @param color: with which to paint C{state}
        @type color: str of valid dot color
        """
        self.graph.node[state]['style'] = 'filled'
        self.graph.node[state]['fillcolor'] = color

    def find(self, states=None, with_attr_dict=None, **with_attr):
        """Filter by desired states and by desired state labels.

        Examples
        ========
        Assume that the system is:

        >>> import transys as trs
        >>> ts = trs.FTS()
        >>> ts.atomic_propositions.add('p')
        >>> ts.states.add('s0', ap={'p'})

          - To find the label of a single state C{'s0'}:

              >>> a = ts.states.find(['s0'] )
              >>> (s0_, label) = a[0]
              >>> print(label)
              {'ap': set(['p'])}

          - To find all states with a specific label C{{'p'}}:

              >>> ts.states.add('s1', ap={'p'})
              >>> b = ts.states.find(with_attr_dict={'ap':{'p'} } )
              >>> states = [state for (state, label_) in b]
              >>> print(set(states) )
              {'s0', 's1'}

          - To find all states in subset C{M} labeled with C{{'p'}}:

              >>> ts.states.add('s2', ap={'p'})
              >>> M = {'s0', 's2'}
              >>> b = ts.states.find(M, {'ap': {'p'} } )
              >>> states = [state for (state, label_) in b]
              >>> print(set(states) )
              {'s0', 's2'}

        @param states: subset of states over which to search
        @type states: 'any' (default)
            | iterable of valid states
            | single valid state

        @param with_attr_dict: label with which to filter the states
        @type with_attr_dict: {sublabel_type : desired_sublabel_value, ...}
            | leave empty, to allow any state label (default)

        @param with_attr: label key-value pairs which take
            precedence over C{with_attr_dict}.

        @rtype: list of labeled states
        @return: [(C{state}, C{label}),...]
            where:
                - C{state} \\in C{states}
                - C{label}: dict
        """
        if with_attr_dict is None:
            with_attr_dict = with_attr
        else:
            try:
                with_attr_dict.update(with_attr)
            except AttributeError:
                raise Exception('with_attr_dict must be a dict')
        if states is not None:
            # singleton check
            if states in self:
                state = states
                msg = (
                    'LabeledStates.find got single state: ' +
                    str(state) + '\n'
                    'instead of Iterable of states.\n')
                states = [state]
                msg += 'Replaced given states = ' + str(state)
                msg += ' with states = ' + str(states)
                logger.debug(msg)
        found_state_label_pairs = []
        for state, attr_dict in self.graph.nodes_iter(data=True):
            logger.debug('Checking state_id = ' + str(state) +
                         ', with attr_dict = ' + str(attr_dict))
            if states is not None:
                if state not in states:
                    logger.debug('state_id = ' + str(state) + ', not desired.')
                    continue
            msg = (
                'Checking state label:\n\t attr_dict = ' +
                str(attr_dict) +
                '\n vs:\n\t desired_label = ' + str(with_attr_dict))
            logger.debug(msg)
            if not with_attr_dict:
                logger.debug('Any label acceptable.')
                ok = True
            else:
                ok = label_is_desired(attr_dict, with_attr_dict)
            if ok:
                logger.debug('Label Matched:\n\t' + str(attr_dict) +
                             ' == ' + str(with_attr_dict))
                state_label_pair = (state, dict(attr_dict))
                found_state_label_pairs.append(state_label_pair)
            else:
                logger.debug('No match for label---> state discarded.')
        return found_state_label_pairs

    def is_terminal(self, state):
        """Return True if state has no outgoing transitions.

        See Also
        ========
        Def. 2.4, p.23 U{[BK08]
        <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}
        """
        return not bool(self.graph.successors(state))


class Transitions(object):
    """Methods for handling labeled transitions.

    Note that a directed edge is an ordered set of nodes.
    Unlike an edge, a transition is a labeled edge.
    """

    def __init__(self, graph, deterministic=False):
        """Initialize C{Transitions}.

        @type graph: L{LabeledDiGraph}
        """
        self.graph = graph
        self._deterministic = deterministic

    def __call__(self, **kwargs):
        """Return list of transitions.

        Wraps L{LabeledDiGraph.edges}.
        """
        return self.graph.edges(**kwargs)

    def __str__(self):
        return 'Transitions:\n' + pformat(self())

    def __len__(self):
        """Count transitions."""
        return self.graph.number_of_edges()

    def _breaks_determinism(self, from_state, sublabels):
        """Return True if adding transition conserves determinism."""
        if not self._deterministic:
            return

        if from_state not in self.graph.states:
            raise Exception('from_state \notin graph')

        same_labeled = self.find([from_state], with_attr_dict=sublabels)

        if same_labeled:
            msg = (
                'Candidate transition violates determinism.\n'
                'Existing transitions with same label:\n' +
                str(same_labeled)
            )
            raise Exception(msg)

    def add(self, from_state, to_state, attr_dict=None, check=True, **attr):
        """Wrapper of L{LabeledDiGraph.add_edge},

        which wraps C{networkx.MultiDiGraph.add_edge}.
        """
        # self._breaks_determinism(from_state, labels)
        self.graph.add_edge(from_state, to_state,
                            attr_dict=attr_dict, check=check, **attr)

    def add_from(self, transitions, attr_dict=None, check=True, **attr):
        """Wrapper of L{LabeledDiGraph.add_edges_from},

        which wraps C{networkx.MultiDiGraph.add_edges_from}.
        """
        self.graph.add_edges_from(transitions, attr_dict=attr_dict,
                                  check=check, **attr)

    def add_comb(self, from_states, to_states, attr_dict=None,
                 check=True, **attr):
        """Add an edge for each combination C{(u, v)},

        for C{u} in C{from_states} for C{v} in C{to_states}.
        """
        for u in from_states:
            for v in to_states:
                self.graph.add_edge(u, v, attr_dict=attr_dict,
                                    check=check, **attr)

    def remove(self, from_state, to_state, attr_dict=None, **attr):
        """Remove single transition.

        If only the states are passed,
        then all transitions between them are removed.

        If C{attr_dict}, C{attr} are also passed,
        then only transitions annotated with those labels are removed.

        Wraps L{LabeledDiGraph.remove_labeled_edge}.
        """
        self.graph.remove_labeled_edge(from_state, to_state, attr_dict, **attr)

    def remove_from(self, transitions):
        """Remove list of transitions.

        Each transition is either a:

          - 2-tuple: (u, v), or a
          - 3-tuple: (u, v, data)
        """
        self.graph.remove_labeled_edges_from(transitions)

    def add_adj(
            self, adj, adj2states, attr_dict=None,
            check=True, **attr
    ):
        """Add multiple labeled transitions from adjacency matrix.

        The label can be empty.
        For more details see L{add}.

        @param adj: new transitions represented by adjacency matrix.
        @type adj: scipy.sparse.lil (list of lists)

        @param adj2states: map from adjacency matrix indices to states.
            If value not a state, raise Exception.
            Use L{States.add}, L{States.add_from} to add states first.

            For example the 1st state in adj2states corresponds to
            the first node in C{adj}.

            States must have been added using:

               - sys.states.add, or
               - sys.states.add_from

            If C{adj2states} includes a state not in sys.states,
            no transition is added and an exception raised.
        @type adj2states: either of:
            - C{dict} from adjacency matrix indices to
              existing, or
            - C{list} of existing states
        """
        # square ?
        if adj.shape[0] != adj.shape[1]:
            raise Exception('Adjacency matrix must be square.')
        # check states exist, before adding any transitions
        for state in adj2states:
            if state not in self.graph:
                raise Exception(
                    'State: ' + str(state) + ' not found.'
                    ' Consider adding it with sys.states.add')
        # convert to format friendly for edge iteration
        nx_adj = nx.from_scipy_sparse_matrix(
            adj, create_using=nx.DiGraph())
        # add each edge using existing checks
        for i, j in nx_adj.edges_iter():
            si = adj2states[i]
            sj = adj2states[j]
            self.add(si, sj, attr_dict, check, **attr)

    def find(self, from_states=None, to_states=None,
             with_attr_dict=None, typed_only=False, **with_attr):
        """Find all edges between given states with given labels.

        Instead of having two separate methods to:

          - find all labels of edges between given states (s1, s2)

          - find all transitions (s1, s2, L) with given label L,
                possibly from some given state s1,
                i.e., the edges leading to the successor states
                Post(s1, a) = Post(s1) restricted by action a

        this method provides both functionalities.

        Preimage under edge labeling function L of given label,
        intersected with given subset of edges::
            L^{-1}(desired_label) \\cap (from_states x to_states)

        See Also
        ========
        L{add}, L{add_adj}

        @param from_states: edges must start from this subset of states
        @type from_states:
            - iterable of existing states, or
            - None (no constraint, default)

        @param to_states: edges must end in this subset of states
        @type to_states:
            - iterable of existing states, or
            - None (no constraint, default)

        @param with_attr_dict: edges must be annotated with these labels
        @type with_attr_dict:
            - {label_type : desired_label_value, ...}, or
            - None (no constraint, default)

        @param with_attr: label type-value pairs,
            take precedence over C{desired_label}.

        @return: set of transitions = labeled edges::
                (from_state, to_state, label)
        such that::
                (from_state, to_state )
                in from_states x to_states

        @rtype: list of transitions::
                = list of labeled edges
                = [(from_state, to_state, label),...]
        where:
          - C{from_state} in C{from_states}
          - C{to_state} in C{to_states}
          - C{label}: dict
        """
        if with_attr_dict is None:
            with_attr_dict = with_attr
        try:
            with_attr_dict.update(with_attr)
        except:
            raise TypeError('with_attr_dict must be a dict')
        found_transitions = []
        u_v_edges = self.graph.edges_iter(nbunch=from_states, data=True)
        if to_states is not None:
            u_v_edges = [(u, v, d)
                         for u, v, d in u_v_edges
                         if v in to_states]
        for u, v, attr_dict in u_v_edges:
            ok = True
            if not with_attr_dict:
                logger.debug('Any label is allowed.')
            elif not attr_dict:
                logger.debug('No labels defined.')
            else:
                logger.debug('Checking guard.')
                ok = label_is_desired(attr_dict, with_attr_dict)
            if ok:
                logger.debug('Transition label matched desired label.')
                transition = (u, v, dict(attr_dict))
                found_transitions.append(transition)
        return found_transitions


class LabeledDiGraph(nx.MultiDiGraph):
    """Directed multi-graph with constrained labeling.

    Provides facilities to define labeling functions on
    vertices and edges, with given co-domains,
    so that labels are type-checked, not arbitrary.

    Each state (or edge) is annotated with labels.
    Before removing a value from a label type,
    first make sure no state (or edge) is labeled with it.

    Multiple edges with the same C{attr_dict} are not possible.
    So the difference from C{networkx.MultiDiGraph} is that
    the C{dict} of edges between u,v is a bijection.

    Between two nodes either:

      - a single unlabeled edge exists (no labeling constraints), or
      - labeled edges exist

    but mixing labeled with unlabeled edges for the same
    edge is not allowed, to simplify and avoid confusion.


    Label types by example
    ======================

    Use a C{dict} for each label type you want to define,
    like this:

      >>> types = [
              {'name': 'drink',
               'values': {'tea', 'coffee'},
               'setter': True,
               'default': 'tea'}]

    This will create a label type named C{'drink'} that can
    take the values C{'tea'} and C{'coffee'}.

    Assuming this label type applies to nodes,
    you can now label a new node as:

      >>> g = LabeledDiGraph(types)
      >>> g.add_node(1, drink='coffee')

    If you omit the label when adding a new node,
    it gets the default value:

      >>> g.add_node(2)
      >>> g.node[2]
      {'drink': 'tea'}

    The main difference with vanilla C{networkx} is
    that the dict above includes type checking:

      >>> type(g.node[2])
      tulip.transys.mathset.TypedDict

    The C{'setter'} key with value C{True}
    creates also a field C{g.drink}.
    Be careful to avoid name conflicts with existing
    networkx C{MultiDiGraph} attributes.

    This allows us to add more values after creating
    the graph:

      >>> g.drink
      {'coffee', 'tea'}
      >>> g.drink.add('water')
      {'coffee', 'tea', 'water'}

    Finally, the graph will prevent us from
    accidentally using an untyped label name,
    by raising an C{AttributeError}:

      >>> g.add_node(3, day='Jan')
      AttributeError: ...

    To add untyped labels, do so explicitly:

      >>> g.add_node(3, day='Jan', check=False)
      >>> g.node[3]
      {'day': 'Jan', 'drink': 'tea'}


    Details on label types
    ======================

    Each label type is defined by a C{dict} that
    must have the keys C{'name'} and C{'values'}:

      - C{'name'}: with C{str} value

      - C{'values' : B} implements C{__contains__}
        used to check label validity.

        If you want the codomain C{B} to be
        extensible even after initialization,
        it must implement method C{add}.

    and optionally the keys:

      - C{'setter': C} with 3 possibilities:

        - if absent,
          then no C{setter} attribute is created

        - otherwise an attribute C{self.A}
          is created, pointing at:

            - the given co-domain C{B}
              if C{C is True}

            - C{C}, otherwise.

      - C{'default': d} is a value in C{B}
        to be returned for node and edge labels
        not yet explicitly specified by the user.

    @param node_label_types: applies to nodes, as described above.
    @type node_label_types: C{list} of C{dict}

    @param edge_label_types: applies to edges, as described above.
    @type node_label_types: C{list} of C{dict}

    @param deterministic: if True, then edge-label-deterministic


    Deprecated dot export
    =====================

    BEWARE: the dot interface will be separated from
    the class itself. Some basic style definitions as
    below may remain, but masking selected labels and
    other features will be accessible only via functions.

    For dot export subclasses must define:

        - _state_label_def
        - _state_dot_label_format

        - _transition_label_def
        - _transition_dot_label_format
        - _transition_dot_mask

    Note: this interface will be improved in the future.


    Credits
    =======

    Some code in overridden methods of C{networkx.MultiDiGraph}
    is adapted from C{networkx}, which is distributed under a BSD license.
    """

    def __init__(
            self,
            node_label_types=None,
            edge_label_types=None,
            deterministic=False):
        node_labeling, node_defaults = self._init_labeling(node_label_types)
        edge_labeling, edge_defaults = self._init_labeling(edge_label_types)

        self._state_label_def = node_labeling
        self._node_label_defaults = node_defaults

        self._transition_label_def = edge_labeling
        self._edge_label_defaults = edge_defaults

        # temporary hack until rename
        self._node_label_types = self._state_label_def
        self._edge_label_types = self._transition_label_def

        nx.MultiDiGraph.__init__(self)

        self.states = States(self)

        # todo: handle accepting states separately
        self.transitions = Transitions(self, deterministic)

        # export properties
        self.dot_node_shape = {'normal': 'circle'}
        self.default_layout = 'dot'

    def _init_labeling(self, label_types):
        """Initialize labeling.

        Note
        ====
        'state' will be renamed to 'node' in the future
        'transition' will be renamed to 'edge' in the future

        @param label_types: see L{__init__}.
        """
        labeling = dict()
        defaults = dict()
        if label_types is None:
            logger.debug('no label types passed')
            return labeling, defaults
        if not label_types:
            logger.warn('empty label types: %s' % str(label_types))
        # define the labeling
        labeling = {d['name']: d['values'] for d in label_types}
        defaults = {d['name']: d.get('default') for d in label_types
                    if 'default' in d}
        setters = {d['name']: d.get('setter') for d in label_types
                   if 'setter' in d}
        for name, setter in setters.iteritems():
            # point to given values ?
            if setter is True:
                setter = labeling[name]
            setattr(self, name, setter)
        return labeling, defaults

    def _check_for_untyped_keys(self, typed_attr, type_defs, check):
        untyped_keys = set(typed_attr).difference(type_defs)
        msg = (
            'checking for untyped keys...\n' +
            'attribute dict: ' + str(typed_attr) + '\n' +
            'type definitions: ' + str(type_defs) + '\n' +
            'untyped_keys: ' + str(untyped_keys))
        logger.debug(msg)
        if untyped_keys:
            msg = (
                'The following edge attributes:\n' +
                str({k: typed_attr[k] for k in untyped_keys}) + '\n' +
                'are not allowed.\n' +
                'Currently the allowed attributes are:' +
                ', '.join([str(x) for x in type_defs]))
            if check:
                msg += ('\nTo set attributes not included ' +
                        'in the existing types, pass: check = False')
                raise AttributeError(msg)
            else:
                msg += '\nAllowed because you passed: check = True'
                logger.warning(msg)
        else:
            logger.debug('no untyped keys.')

    def is_consistent(self):
        """Check if labels are consistent with their type definitions.

        Use case: removing values from a label type
        can invalidate existing labels that use them.

        @rtype: bool
        """
        for node, attr_dict in self.nodes_iter(data=True):
            if not attr_dict.is_consistent():
                return False
        for node_i, node_j, attr_dict in self.edges_iter(data=True):
            if not attr_dict.is_consistent():
                return False
        return True

    def _update_attr_dict_with_attr(self, attr_dict, attr):
        if attr_dict is None:
            attr_dict = attr
        else:
            try:
                attr_dict.update(attr)
            except AttributeError:
                msg = 'The attr_dict argument must be a dictionary.'
                raise nx.NetworkXError(msg)
        return attr_dict

    def add_node(self, n, attr_dict=None, check=True, **attr):
        """Use a L{TypedDict} as attribute dict.

        Overrides C{networkx.MultiDiGraph.add_node},
        see that for details.

        Log warning if node already exists.
        All other functionality remains the same.

        @param check: if True and untyped keys are passed,
            then raise C{AttributeError}.
        """
        # avoid multiple additions
        if n in self:
            logger.debug('Graph already has node: ' + str(n))
        attr_dict = self._update_attr_dict_with_attr(attr_dict, attr)
        # define typed dict
        typed_attr = TypedDict()
        typed_attr.set_types(self._node_label_types)
        typed_attr.update(copy.deepcopy(self._node_label_defaults))
        # type checking happens here
        typed_attr.update(attr_dict)
        logger.debug('node typed_attr: ' + str(typed_attr))
        self._check_for_untyped_keys(typed_attr,
                                     self._node_label_types,
                                     check)
        nx.MultiDiGraph.add_node(self, n, attr_dict=typed_attr)

    def add_nodes_from(self, nodes, check=True, **attr):
        """Create or label multiple nodes.

        Overrides C{networkx.MultiDiGraph.add_nodes_from},
        for details see that and L{LabeledDiGraph.add_node}.
        """
        for n in nodes:
            try:
                n not in self.succ
                node = n
                attr_dict = attr
            except TypeError:
                node, ndict = n
                attr_dict = attr.copy()
                attr_dict.update(ndict)
            self.add_node(node, attr_dict=attr_dict, check=check)

    def add_edge(self, u, v, key=None, attr_dict=None, check=True, **attr):
        """Use a L{TypedDict} as attribute dict.

        Overrides C{networkx.MultiDiGraph.add_edge},
        see that for details.

          - Raise ValueError if C{u} or C{v} are not already nodes.
          - Raise Exception if edge (u, v, {}) exists.
          - Log warning if edge (u, v, attr_dict) exists.
          - Raise ValueError if C{attr_dict} contains
            typed key with invalid value.
          - Raise AttributeError if C{attr_dict} contains untyped keys,
            unless C{check=False}.

        Each label defines a different labeled edge.
        So to "change" the label, either:

            - remove the edge with this label, then add a new one, or
            - find the edge key, then use subscript notation:

                C{G[i][j][key]['attr_name'] = attr_value}

        Notes
        =====
        1. Argument C{key} has been removed compared to
           C{networkx.MultiDiGraph.add_edge}, because edges are defined
           by their labeling, i.e., multiple edges with same labeling
           are not allowed.

        @param check: raise C{AttributeError} if C{attr_dict}
            has untyped attribute keys, otherwise warn
        """
        # legacy
        if 'check_states' in attr:
            msg = 'saw keyword argument: check_states ' +\
                  'which is no longer available, ' +\
                  'firstly add the new nodes.'
            logger.warning(msg)
        # check nodes exist
        if u not in self.succ:
            raise ValueError('Graph does not have node u: ' + str(u))
        if v not in self.succ:
            raise ValueError('Graph does not have node v: ' + str(v))
        attr_dict = self._update_attr_dict_with_attr(attr_dict, attr)
        # define typed dict
        typed_attr = TypedDict()
        typed_attr.set_types(self._edge_label_types)
        typed_attr.update(copy.deepcopy(self._edge_label_defaults))
        # type checking happens here
        typed_attr.update(attr_dict)
        logger.debug('Given: attr_dict = ' + str(attr_dict))
        logger.debug('Stored in: typed_attr = ' + str(typed_attr))
        # may be possible to speedup using .succ
        existing_u_v = self.get_edge_data(u, v, default={})
        if dict() in existing_u_v.values():
            msg = (
                'Unlabeled transition: '
                'from_state-> to_state already exists,\n'
                'where:\t from_state = ' + str(u) + '\n'
                'and:\t to_state = ' + str(v) + '\n')
            raise Exception(msg)
        # check if same labeled transition exists
        if attr_dict in existing_u_v.values():
            msg = (
                'Same labeled transition:\n'
                'from_state---[label]---> to_state\n'
                'already exists, where:\n'
                '\t from_state = ' + str(u) + '\n'
                '\t to_state = ' + str(v) + '\n'
                '\t label = ' + str(typed_attr) + '\n')
            warnings.warn(msg)
            logger.warning(msg)
            return
        # self._breaks_determinism(from_state, labels)
        self._check_for_untyped_keys(typed_attr,
                                     self._edge_label_types,
                                     check)
        # the only change from nx in this clause is using TypedDict
        logger.debug('adding edge: ' + str(u) + ' ---> ' + str(v))
        if v in self.succ[u]:
            msg = 'there already exist directed edges with ' +\
                  'same end-points'
            logger.debug(msg)
            keydict = self.adj[u][v]
            # find a unique integer key
            if key is None:
                key = len(keydict)
                while key in keydict:
                    key -= 1
            datadict = keydict.get(key, typed_attr)
            datadict.update(typed_attr)
            keydict[key] = datadict
        else:
            logger.debug('first directed edge between these nodes')
            # selfloops work this way without special treatment
            key = 0
            keydict = {key: typed_attr}
            self.succ[u][v] = keydict
            self.pred[v][u] = keydict

    def add_edges_from(self, labeled_ebunch, attr_dict=None,
                       check=True, **attr):
        """Add multiple labeled edges.

        Overrides C{networkx.MultiDiGraph.add_edges_from},
        see that for details.

        Only difference is that only 2 and 3-tuple edges allowed.
        Keys cannot be specified, because a bijection is maintained.

        @param labeled_ebunch: iterable container of:

            - 2-tuples: (u, v), or
            - 3-tuples: (u, v, label)

          See also L{remove_labeled_edges_from}.
        """
        attr_dict = self._update_attr_dict_with_attr(attr_dict, attr)
        # process ebunch
        for e in labeled_ebunch:
            datadict = dict(attr_dict)
            # distinguish by number of elements given
            ne = len(e)
            if ne == 4:
                u, v, key, dd = e
            elif ne == 3:
                u, v, dd = e
                key = None
            elif ne == 2:
                u, v = e
                dd = {}
                key = None
            else:
                raise ValueError(
                    'Edge tuple %s must be a 2-, 3-, or 4-tuple .' % (e,))
            datadict.update(dd)
            self.add_edge(u, v, key=key, attr_dict=datadict, check=check)

    def remove_labeled_edge(self, u, v, attr_dict=None, **attr):
        """Remove single labeled edge.

        @param attr_dict: attributes with which to identify the edge.
        @type attr_dict: dict

        @param attr: keyword arguments with which to update C{attr_dict}.
        """
        if u not in self:
            return
        if v not in self[u]:
            return
        attr_dict = self._update_attr_dict_with_attr(attr_dict, attr)
        rm_keys = {key for key, data in self[u][v].iteritems()
                   if data == attr_dict}
        for key in rm_keys:
            self.remove_edge(u, v, key=key)

    def remove_labeled_edges_from(self, labeled_ebunch,
                                  attr_dict=None, **attr):
        """Remove labeled edges.

        Example
        =======
        >>> g = LabeledDiGraph()
        >>> g.add_edge(1, 2, day='Mon')
        >>> g.add_edge(1, 2, day='Tue')
        >>> edges = [(1, 2, {'day':'Mon'}),
                     (1, 2, {'day':'Tue'})]
        >>> g.remove_edges_from(edges)

        @param labeled_ebunch: iterable container of edge tuples
            Each edge tuple can be:

              - 2-tuple: (u, v) All edges between u and v are removed.
              - 3-tuple: (u, v, attr_dict) all edges between u and v
                  annotated with that C{attr_dict} are removed.
        """
        attr_dict = self._update_attr_dict_with_attr(attr_dict, attr)
        for e in labeled_ebunch:
            datadict = dict(attr_dict)
            ne = len(e)
            if ne == 3:
                u, v, dd = e
                datadict.update(dd)
            elif ne == 2:
                u, v = e
            else:
                raise ValueError(
                    'Edge tuple %s must be a 2- or 3-tuple .' % (e,))
            self.remove_labeled_edge(u, v, attr_dict=datadict)

    def has_deadends(self):
        """Return False if all nodes have outgoing edges.

        Edge labels are not taken into account.
        """
        for n in self:
            if not self.succ[n]:
                return True
        return False

    def remove_deadends(self):
        """Recursively delete nodes with no outgoing transitions."""
        n = len(self)
        s = {1}
        while s:
            s = {n for n in self if not self.succ[n]}
            self.states.remove_from(s)
        m = len(self)
        assert n == 0 or m > 0, 'removed all {n} nodes!'.format(n=n)
        assert n >= 0, 'added {n} nodes'.format(n=n)
        print('removed {r} nodes from '
              '{n} total'.format(r=n - m, n=n))

    def dot_str(self, wrap=10, **kwargs):
        """Return dot string.

        Requires pydot.
        """
        from tulip.transys.export import graph2dot
        return graph2dot.graph2dot_str(self, wrap, **kwargs)

    def save(self, filename=None, fileformat=None,
             rankdir='LR', prog=None,
             wrap=10, tikz=False):
        """Save image to file.

        Recommended file formats:

            - tikz (via dot2tex)
            - pdf
            - svg
            - dot
            - png

        Any other format supported by C{pydot.write} is available.

        Experimental:

            - html (uses d3.js)
            - 'scxml'

        Requires
        ========
          - graphviz dot: http://www.graphviz.org/
          - pydot: https://pypi.python.org/pypi/pydot

        and for tikz:

          - dot2tex: https://pypi.python.org/pypi/dot2tex
          - dot2texi: http://www.ctan.org/pkg/dot2texi
            (to automate inclusion)

        See Also
        ========
        L{plot}, C{pydot.Dot.write}

        @param filename: file path to save image to
            Default is C{self.name}, unless C{name} is empty,
            then use 'out.pdf'.

            If extension is missing '.pdf' is used.
        @type filename: str

        @param fileformat: replace the extension of C{filename}
            with this. For example::

                filename = 'fig.pdf'
                fileformat = 'svg'

            result in saving 'fig.svg'

        @param rankdir: direction for dot layout
        @type rankdir: str = 'TB' | 'LR'
            (i.e., Top->Bottom | Left->Right)

        @param prog: executable to call
        @type prog: dot | circo | ... see pydot.Dot.write

        @param wrap: max width of node strings
        @type wrap: int

        @param tikz: use tikz automata library in dot

        @rtype: bool
        @return: True if saving completed successfully, False otherwise.
        """
        if filename is None:
            if not self.name:
                filename = 'out'
            else:
                filename = self.name
        fname, fextension = os.path.splitext(filename)
        # default extension
        if not fextension or fextension is '.':
            fextension = '.pdf'
        if fileformat:
            fextension = '.' + fileformat
        filename = fname + fextension
        # drop '.'
        fileformat = fextension[1:]
        # check for html
        if fileformat is 'html':
            from tulip.transys.export import save_d3
            return save_d3.labeled_digraph2d3(self, filename)
        # subclass has extra export formats ?
        if hasattr(self, '_save'):
            if self._save(filename, fileformat):
                return True
        if prog is None:
            prog = self.default_layout
        from tulip.transys.export import graph2dot
        graph2dot.save_dot(self, filename, fileformat, rankdir,
                           prog, wrap, tikz=tikz)
        return True

    def plot(self, rankdir='LR', prog=None, wrap=10, ax=None):
        """Plot image using dot.

        No file I/O involved.
        Requires GraphViz dot and either Matplotlib or IPython.

        NetworkX does not yet support plotting multiple edges between 2 nodes.
        This method fixes that issue, so users don't need to look at files
        in a separate viewer during development.

        See Also
        ========
        L{save}

        Depends
        =======
        dot and either of IPython or Matplotlib
        """
        # anything to plot ?
        if not self.states:
            print(
                60 * '!' +
                "\nThe system doesn't have any states to plot.\n" +
                60 * '!')
            return
        if prog is None:
            prog = self.default_layout
        from tulip.transys.export import graph2dot
        return graph2dot.plot_pydot(self, prog, rankdir, wrap, ax=ax)


def str2singleton(ap_label):
    """If string, convert to set(string).

    Convention: singleton str {'*'}
    can be passed as str '*' instead.
    """
    if isinstance(ap_label, str):
        logger.debug('Saw str state label:\n\t' + str(ap_label))
        ap_label = {ap_label}
        logger.debug('Replaced with singleton:\n\t' + str(ap_label) + '\n')
    return ap_label


def prepend_with(states, prepend_str):
    """Prepend items with given string.

    Example
    =======
    >>> states = [0, 1]
    >>> prepend_str = 's'
    >>> states = prepend_with(states, prepend_str)
    >>> assert(states == ['s0', 's1'] )

    See Also
    ========
    L{tuple2ba}, L{tuple2fts}

    @param states: items prepended with string C{prepend_str}
    @type states: iterable

    @param prepend_str: text prepended to C{states}.  If None, then
        C{states} is returned without modification
    @type prepend_str: str or None
    """
    if not isinstance(states, Iterable):
        raise TypeError('states must be Iterable. Got:\n\t' +
                        str(states) + '\ninstead.')
    if not isinstance(prepend_str, str) and prepend_str is not None:
        raise TypeError('prepend_str must be of type str. Got:\n\t' +
                        str(prepend_str) + '\ninstead.')
    if prepend_str is None:
        return states
    return [prepend_str + str(s) for s in states]



#!/usr/bin/env python
"""Controller synthesis for system with continuous (linear) dynamics.

This example is an extension of `robot_discrete.py`,
by including continuous dynamics with disturbances.
The dynamics is linear over a bounded set that is a polytope.
"""
# Petter Nilsson (pettni@kth.se)
# August 14, 2011
# NO, system and cont. prop definitions based on TuLiP 1.x
# 2 Jul, 2013
# NO, TuLiP 1.x discretization
# 17 Jul, 2013

# Note: This code is commented to allow components to be extracted into
# the tutorial that is part of the users manual.  Comments containing
# strings of the form @label@ are used for this purpose.

# @import_section@
import logging

import numpy as np
from tulip import spec, synth, hybrid
from polytope import box2poly
from tulip.abstract import prop2part, discretize
from tulip.abstract.plot import plot_partition
# @import_section_end@


logging.basicConfig(level=logging.WARNING)
show = False

# @dynamics_section@
# Problem parameters
input_bound = 1.0
uncertainty = 0.01

# Continuous state space
cont_state_space = box2poly([[0., 3.], [0., 2.]])

# Continuous dynamics
A = np.array([[1.0, 0.], [ 0., 1.0]])
B = np.array([[0.1, 0.], [ 0., 0.1]])
E = np.array([[1,0], [0,1]])

# Available control, possible disturbances
U = input_bound *np.array([[-1., 1.], [-1., 1.]])
W = uncertainty *np.array([[-1., 1.], [-1., 1.]])

# Convert to polyhedral representation
U = box2poly(U)
W = box2poly(W)

# Construct the LTI system describing the dynamics
sys_dyn = hybrid.LtiSysDyn(A, B, E, None, U, W, cont_state_space)
# @dynamics_section_end@

# @partition_section@
# Define atomic propositions for relevant regions of state space
cont_props = {}
cont_props['home'] = box2poly([[0., 1.], [0., 1.]])
cont_props['lot'] = box2poly([[2., 3.], [1., 2.]])

# Compute the proposition preserving partition of the continuous state space
cont_partition = prop2part(cont_state_space, cont_props)
plot_partition(cont_partition) if show else None
# @partition_section_end@

# @discretize_section@
# Given dynamics & proposition-preserving partition, find feasible transitions
disc_dynamics = discretize(
    cont_partition, sys_dyn, closed_loop=True,
    N=8, min_cell_volume=0.1, plotit=show
)
# @discretize_section_end@

# Visualize transitions in continuous domain (optional)
plot_partition(disc_dynamics.ppp, disc_dynamics.ts,
               disc_dynamics.ppp2ts) if show else None

# Specifications
# Environment variables and assumptions
env_vars = {'park'}
env_init = set()                # empty set
env_prog = '!park'
env_safe = set()                # empty set

# System variables and requirements
sys_vars = {'X0reach'}
sys_init = {'X0reach'}
sys_prog = {'home'}               # []<>home
sys_safe = {'(X(X0reach) <-> lot) || (X0reach && !park)'}
sys_prog |= {'X0reach'}

# Create the specification
specs = spec.GRSpec(env_vars, sys_vars, env_init, sys_init,
                    env_safe, sys_safe, env_prog, sys_prog)
specs.qinit = '\E \A'

# @synthesize_section@
# Synthesize
ctrl = synth.synthesize('omega', specs,
                        sys=disc_dynamics.ts, ignore_sys_init=True)
assert ctrl is not None, 'unrealizable'


# Generate a graphical representation of the controller for viewing
if not ctrl.save('continuous.png'):
    print(ctrl)
# @synthesize_section_end@

# Simulation
